/*
COPYRIGHT (C) 2001 Guennadi Liakhovetski, DSA GmbH (gl@dsa-ac.de)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/

#ifndef _ASM_IRQOPS_H
#define _ASM_IRQOPS_H

//#define NR_SRQS NR_GLOBAL_IRQS

//static unsigned int pending_srqs[NR_SYSRQS];

#define INC_WRAP(x,y) do { x = ++(x) & ((y)-1); } while (0)

static inline int llffnz( unsigned long long ull )
{
	return ((unsigned long *)&ull)[0] ?
		ffnz( ((unsigned long *)&ull)[0] ) :
		( ((unsigned long *)&ull)[1] ?
		  ffnz( ((unsigned long *)&ull)[1] ) + 32 :
		  -1 );
}

static inline int have_pending_irq( void )
{
	return ( global.pending_irqs[global.irq_out] != -1 );
//	return ( global.pending_irqs != 0 );
}

static inline int have_pending_srq( void )
{
	return ( global.pending_srqs != 0 );
/*	unsigned int i;
	for ( i = 0; i < NR_SYSRQS; i++ )
		if ( global.pending_srqs[i] )
			return 1;
	return 0;*/
}

static inline rtai_irq_t pending_irq( void )
{
	return global.pending_irqs[global.irq_out];
//	return llffnz(global.pending_irqs & ~global.active_irqs);
//	return ffnz(global.pending_irqs & ~global.active_irqs);
}

static inline rtai_irq_t pending_srq( void )
{
	return llffnz(global.pending_srqs & ~global.active_srqs);
/*	int i;

	for ( i = 0; i < NR_SYSRQS; i++ )
		if ( global.pending_srqs[i] && ! test_bit( i, (void *) &global.active_srqs ) )
			return i;
	return -1;*/
}

static inline void activate_irq( rtai_irq_t irq )
{
	set_bit(irq, &global.active_irqs);
}

static inline void deactivate_irq( rtai_irq_t irq )
{
	clear_bit(irq, &global.active_irqs);
}

void rt_pend_linux_irq(rtai_irq_t irq)
{
	if ( global.pending_irqs[global.irq_in] != -1 ) {
		global.lost_irqs++;
		rt_printk( "Lost irq: %d! Increase MAX_PENDING_IRQS (currently %d)\n",
			   global.pending_irqs[global.irq_in], MAX_PENDING_IRQS );
	} //else {
		global.pending_irqs[global.irq_in] = irq;
		INC_WRAP( global.irq_in, MAX_PENDING_IRQS );
	//}
//	set_bit(irq, &global.pending_irqs);
}

static inline void clear_pending_irq( rtai_irq_t irq )
{
	global.pending_irqs[global.irq_out] = -1;
	INC_WRAP( global.irq_out, MAX_PENDING_IRQS );
//	clear_bit(irq, &global.pending_irqs);
}

static inline void activate_srq( rtai_irq_t srq )
{
	set_bit(srq, &global.active_srqs);
}

static inline void deactivate_srq( rtai_irq_t srq )
{
	clear_bit(srq, &global.active_srqs);
}

void rt_pend_linux_srq(rtai_irq_t srq)
{
	if ( srq < 0 || srq >= NR_SYSRQS ) {
		printk( "Invalid SRQ! (pend:) %d\n", srq );
		return;
	}
	set_bit(srq, &global.pending_srqs);
//	global.pending_srqs[srq]++;
}

static inline void clear_pending_srq( rtai_irq_t srq )
{
	if ( srq < 0 || srq >= NR_SYSRQS ) {
		printk( "Invalid SRQ! (clear:) %d\n", srq );
		return;
	}
	clear_bit(srq, &global.pending_srqs);
//	global.pending_srqs[srq]--;
}

static inline void init_pending_srqs( void )
{
	global.pending_srqs = global.active_srqs = 0;
//	global.pending_srqs = pending_srqs;
//	memset( global.pending_srqs, 0, NR_SYSRQS * sizeof( *pending_srqs ) );
}

static inline void init_pending_irqs( void )
{
	int i;

	global.irq_in = global.irq_out = global.lost_irqs = 0;
	for ( i = 0; i < MAX_PENDING_IRQS; i++ )
		global.pending_irqs[i] = -1;
//	global.pending_irqs = 0;
}

#endif
