/*
COPYRIGHT (C) 2000  Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/


#include <stdio.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

#include <asm/io.h>

#define  KEEP_STATIC_INLINE
#include <rtai_lxrt_user.h>
#include <rtai_lxrt.h>

int main(void)
{
	char s[1000];
	int i, fd;
	long long ll;
	float f;
	double d;
	RT_TASK *mytask;
	RTIME t;
        struct timeval timout;

	struct sched_param mysched;

	mysched.sched_priority = sched_get_priority_max(SCHED_FIFO) - 1;
	if( sched_setscheduler( 0, SCHED_FIFO, &mysched ) == -1 ) {
		puts("ERROR IN SETTING THE SCHEDULER");
		perror("errno");
		exit(0);
 	}       
 	if (!(mytask = rt_task_init(nam2num("HRTSK"), 2, 0, 0))) {
		printf("CANNOT INIT TEST BUDDY TASK\n");
		exit(1);
	}
	start_rt_timer(nano2count(1000000));
	rt_grow_and_lock_stack(40000);

	rt_make_hard_real_time();
	rt_task_use_fpu(mytask, 1);

	printf("Input a: string, integer, long long, float, double: ");
	scanf("%s %i %lld %f %lf", s, &i, &ll, &f, &d);
	printf("Got string, integer, long long, now we wait 1 s using select.\n");
        timout.tv_sec = 1;
        timout.tv_usec = 0;
	select(1, NULL, NULL, NULL, &timout);
	printf("Select expired and we print what we read: %s %i %lld %f %f\n", s, i, ll, f, d);
	printf("Open a file.\n");
	fd = open("rtfile", O_RDWR | O_CREAT | O_TRUNC, 0666);
	printf("Write a 50 bytes header of 'B's.\n");
	memset(s, 'B', 50);
	write(fd, s, 50);
	memset(s, 'A', 1000);
	printf("Write 10 KB of 'A's to the opened file.\n");
	t = rt_get_cpu_time_ns();
	for (i = 0; i < 10; i++) {
		write(fd, s, 1000);
		rt_sleep(nano2count(10000000));
	}
	printf("WRITE TIME: %lld (ms).\n", (rt_get_cpu_time_ns() - t + 500000)/1000000);
	printf("Write a 50 bytes trailer of 'E's.\n");
	memset(s, 'E', 50);
	write(fd, s, 50);
	printf("Sync file to disk.\n");
	sync();
	printf("Position file at its beginning.\n");
	lseek(fd, 0, SEEK_SET);
	printf("Read the first 51 bytes (header plus + 'A').\n");
	read(fd, s, 51);
	s[51] = 0;
	printf("Here is the header  %s.\n", s);
	lseek(fd, -1, SEEK_CUR);
	printf("Read the written 10 KB of 'A's back.\n");
	t = rt_get_cpu_time_ns();
	for (i = 0; i < 10; i++) {
		read(fd, s, 1000);
		rt_sleep(nano2count(10000000));
	}
	printf("READ TIME %lld (ms).\n", (rt_get_cpu_time_ns() - t + 500000)/1000000);
	lseek(fd, -1, SEEK_CUR);
	printf("Read the last 51 bytes (last 'A' + trailer).\n");
	read(fd, s, 51);
	s[51] = 0;
	printf("Here is the trailer %s.\n", s);
	printf("Close the file and end the test.\n");
	close(fd);
	rt_make_soft_real_time();
	stop_rt_timer();
	rt_task_delete(mytask);

	return 0;
}
