/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include <FL/fl_ask.H>
#include <FL/Fl_Box.H>
#include <FL/filename.H>
#include "controlpanel.h"
#include "targetpanel.h"
#include "parameterspanel.h"
#include "scopepanel.h"
#include "logpanel.h"
#include "rtailab.h"
#include <pthread.h>

extern Fl_Preferences app;
extern TargetPanel *TargetIDs;
extern ParametersPanel *ParametersManager;
extern ScopePanel *ScopeManager;
extern LogPanel *LogManager;

extern int EndApp;
extern int EndAll;
extern RT_TASK *rt_TargetInterfaceTask; 
extern pthread_t rt_TargetInterfaceThread;
extern pthread_t rt_GetScopeDataThread[];
extern pthread_t rt_GetLogDataThread[];
extern pthread_t rt_GetTargetTimeThread;
extern int NumRtScopes;
extern int NumRtLogs;

extern void cb_connect(Fl_Widget *, void *);
extern void cb_start(Fl_Widget *, void *);

void ControlPanel::show()
{
	Window->show();
}

void ControlPanel::hide()
{
	Window->hide();
}

void ControlPanel::activate_Connect()
{
	Connect->activate();
}

void ControlPanel::activate_Start()
{
	Start->activate();
}

void ControlPanel::activate_SelectTarget()
{
	SelectTarget->activate();
}

void ControlPanel::activate_ChangeParams()
{
	ChangeParams->activate();
}

void ControlPanel::activate_RtPlots()
{
	RtPlots->activate();
}

void ControlPanel::activate_RtLogs()
{
	RtLogs->activate();
}

void ControlPanel::deactivate_Connect()
{
	Connect->deactivate();
}

void ControlPanel::deactivate_Start()
{
	Start->deactivate();
}

void ControlPanel::deactivate_SelectTarget()
{
	SelectTarget->deactivate();
}

void ControlPanel::deactivate_ChangeParams()
{
	ChangeParams->deactivate();
}

void ControlPanel::deactivate_RtPlots()
{
	RtPlots->deactivate();
}

void ControlPanel::deactivate_RtLogs()
{
	RtLogs->deactivate();
}

void ControlPanel::ConnectLabel(const char *buf)
{
	Connect->label(buf);
}

void ControlPanel::StartLabel(const char *buf)
{
	Start->label(buf);
}

void ControlPanel::MessageValue(const char *buf)
{
	Message->value(buf);
}

inline void ControlPanel::cb_seltarget_i(Fl_Button *o, void *v)
{
	char path[FL_PATH_MAX];
	app.getUserdataPath(path, sizeof(path));

	if (!TargetIDs) {
		TargetIDs = new TargetPanel();
	}
	TargetIDs->show();
	SelectTarget->deactivate();
}

void ControlPanel::cb_seltarget(Fl_Button *o, void *v)
{
	((ControlPanel *)(o->parent()->parent()->parent()->user_data()))->cb_seltarget_i(o,v);
}

#ifdef PASSWD

const char *ChgParamPasswd = "" PASSWD "";

inline void ControlPanel::cb_checkpasswd_i(Fl_Secret_Input *o, void *v)
{
	if (!strcmp(ChgParamPasswd, strdup(o->value()))) {
		if (!ParametersManager) {
			ParametersManager = new ParametersPanel();
		}
		ParametersManager->show();
		ChangeParams->deactivate();
	} else {
		fl_alert("Permission denied\n");
	}
	PasswdWindow->hide();
}

void ControlPanel::cb_checkpasswd(Fl_Secret_Input *o, void *v)
{
	((ControlPanel *)(o->parent()->parent()->parent()->user_data()))->cb_checkpasswd_i(o,v);
}

inline void ControlPanel::cb_askpasswd_i(Fl_Button *o, void *v)
{
	if (!PasswdWindow) {
		{ Fl_Window *o = PasswdWindow = new Fl_Window(300, 60);
		  o->label("Private Access");
		  o->user_data((void *)(this));
		  { Fl_Group *o = new Fl_Group(0, 0, 300, 60);
		    { Fl_Group *o = PasswdWidgets = new Fl_Group(0, 0, 300, 60);
		      { Fl_Secret_Input *o = PasswdEntry = new Fl_Secret_Input(100, 20, 150, 20, "Enter password:");
			o->labelsize(12);
			o->align(FL_ALIGN_LEFT);
			o->when(FL_WHEN_ENTER_KEY);
			o->callback((Fl_Callback *)cb_checkpasswd);
		      }
		      o->end();
		    }
		    o->end();
		  }
		}
	} else {
		PasswdEntry->value("");
	}
	PasswdWindow->show();
}

void ControlPanel::cb_askpasswd(Fl_Button *o, void *v)
{
	((ControlPanel *)(o->parent()->parent()->parent()->user_data()))->cb_askpasswd_i(o,v);
}

#else

inline void ControlPanel::cb_changeparams_i(Fl_Button *o, void *v)
{
	if (!ParametersManager) {
		ParametersManager = new ParametersPanel();
	}
	ParametersManager->show();
	ChangeParams->deactivate();
}

void ControlPanel::cb_changeparams(Fl_Button *o, void *v)
{
	((ControlPanel *)(o->parent()->parent()->parent()->user_data()))->cb_changeparams_i(o,v);
}

#endif

inline void ControlPanel::cb_scopes_i(Fl_Button *o, void *v)
{
	if (!ScopeManager) {
		ScopeManager = new ScopePanel();
	}
	ScopeManager->show();
	RtPlots->deactivate();
}

void ControlPanel::cb_scopes(Fl_Button *o, void *v)
{
	((ControlPanel *)(o->parent()->parent()->parent()->user_data()))->cb_scopes_i(o,v);
}

inline void ControlPanel::cb_logs_i(Fl_Button *o, void *v)
{
	if (!LogManager) {
		LogManager = new LogPanel();
	}
	LogManager->show();
	RtLogs->deactivate();
}

void ControlPanel::cb_logs(Fl_Button *o, void *v)
{
	((ControlPanel *)(o->parent()->parent()->parent()->user_data()))->cb_logs_i(o,v);
}

inline void ControlPanel::cb_close_i(Fl_Button *o, void *v)
{
	int n;

	EndApp = 1;
	for (n = 0; n < NumRtScopes; n++) {
		pthread_join(rt_GetScopeDataThread[n], NULL);
	}
	for (n = 0; n < NumRtLogs; n++) {
		pthread_join(rt_GetLogDataThread[n], NULL);
	}
	pthread_join(rt_GetTargetTimeThread, NULL);
	rt_send(rt_TargetInterfaceTask, CLOSE);
	pthread_join(rt_TargetInterfaceThread, NULL);
	EndAll = 1;
}

void ControlPanel::cb_close(Fl_Button *o, void *v)
{
	((ControlPanel *)(o->parent()->parent()->parent()->user_data()))->cb_close_i(o,v);
}

ControlPanel::ControlPanel()
{
	Fl_Window *w;
	{ Fl_Window *o = Window = new Fl_Window(500, 190);
	  w = o;
	  o->label("RTAI-Lab Control Panel");
	  o->user_data((void *)(this));
	  { Fl_Group *o = new Fl_Group(0, 0, 500, 190);
	    { Fl_Group *o = ControlButtons = new Fl_Group(0, 0, 500, 190);
	      { Fl_Box *box = new Fl_Box(FL_FRAME_BOX, 10, 10, 480, 55, 0);
		box->color(FL_GRAY);
	      }
	      { Fl_Box *box = new Fl_Box(FL_FRAME_BOX, 10, 80, 480, 55, 0);
		box->color(FL_GRAY);
	      }
	      { Fl_Button *o = Connect = new Fl_Button(180, 20, 140, 35, "Connect");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_connect);
	      }
	      { Fl_Button *o = Start = new Fl_Button(340, 20, 140, 35, "Start real-time code");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_start);
		o->deactivate();
	      }
	      { Fl_Button *o = SelectTarget = new Fl_Button(20, 20, 140, 35, "Select Target");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_seltarget);
	      }
	      { Fl_Button *o = ChangeParams = new Fl_Button(20, 90, 140, 35, "Change parameters");
		o->labelsize(12);
		o->deactivate();
#ifdef PASSWD
		o->callback((Fl_Callback *)cb_askpasswd);
#else
		o->callback((Fl_Callback *)cb_changeparams);
#endif
	      }
	      { Fl_Button *o = RtPlots = new Fl_Button(180, 90, 140, 35, "Real-time plots");
		o->labelsize(12);
		o->deactivate();
		o->callback((Fl_Callback *)cb_scopes);
	      }
	      { Fl_Button *o = RtLogs = new Fl_Button(340, 90, 140, 35, "Real-time logs");
		o->labelsize(12);
		o->deactivate();
		o->callback((Fl_Callback *)cb_logs);
	      }
	      { Fl_Output *o = Message = new Fl_Output(10, 150, 350, 25, 0);
		o->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
		o->textfont(FL_TIMES_ITALIC);
		o->value(" Not connected");
	      }
	      { Fl_Button *o = Close = new Fl_Button(420, 150, 70, 25, "Close");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_close);
	      }
	      o->end();
	    }
	    o->end();
          }
	}
};
