/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#ifndef _CONTROLPANEL_H
#define _CONTROLPANEL_H

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Secret_Input.H>

class ControlPanel
{
	public:
		ControlPanel();
		void show();
		void hide();
		void activate_Connect();
		void activate_Start();
		void activate_SelectTarget();
		void activate_ChangeParams();
		void activate_RtPlots();
		void activate_RtLogs();
		void deactivate_Connect();
		void deactivate_Start();
		void deactivate_SelectTarget();
		void deactivate_ChangeParams();
		void deactivate_RtPlots();
		void deactivate_RtLogs();
		void ConnectLabel(const char *);
		void StartLabel(const char *);
		void MessageValue(const char *);
	private:
		Fl_Window *Window;
		Fl_Group *ControlButtons;
		Fl_Button *Connect;
		Fl_Button *Start;
		Fl_Button *SelectTarget;
		Fl_Button *ChangeParams;
		Fl_Button *RtPlots;
		Fl_Button *RtLogs;
		Fl_Button *Close;
		Fl_Output *Message;
#ifdef PASSWD
		Fl_Window *PasswdWindow;
		Fl_Group *PasswdWidgets;
		Fl_Secret_Input *PasswdEntry;
#endif
		inline void cb_seltarget_i(Fl_Button *, void *);
		static void cb_seltarget(Fl_Button *, void *);
#ifdef PASSWD
		inline void cb_checkpasswd_i(Fl_Secret_Input *, void *);
		static void cb_checkpasswd(Fl_Secret_Input *, void *);
		inline void cb_askpasswd_i(Fl_Button *, void *);
		static void cb_askpasswd(Fl_Button *, void *);
#else
		inline void cb_changeparams_i(Fl_Button *, void *);
		static void cb_changeparams(Fl_Button *, void *);
#endif
		inline void cb_scopes_i(Fl_Button *, void *);
		static void cb_scopes(Fl_Button *, void *);
		inline void cb_logs_i(Fl_Button *, void *);
		static void cb_logs(Fl_Button *, void *);
		inline void cb_close_i(Fl_Button *, void *);
		static void cb_close(Fl_Button *, void *);
};

#endif
