/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include <FL/forms.H>
#include <FL/Fl_Button.H>
#include "parameterspanel.h"
#include "controlpanel.h"
#include "rtailab.h"
#include <string.h>

extern int NumTunableParameters;
extern rtTargetParamInfo *TunableParameters;
extern ControlPanel *CommandManager;

extern void cb_update(Fl_Widget *, void *);
extern void cb_batch_update(Fl_Widget *, void *);
extern void cb_get_parameters(Fl_Widget *, void *);

void ParametersPanel::show()
{
        Window->show();
}

void ParametersPanel::hide()
{
        Window->hide();
}

int ParametersPanel::IsBatch()
{
	return (int)(BatchDownload->value());
}

int ParametersPanel::BatchCounter()
{
	return counter;
}

int ParametersPanel::increment_BatchCounter()
{
	return counter++;
}

void ParametersPanel::LogTimeAndNames(int idx, struct tm *loctime)
{
	fprintf(LogFile, "Time       : %s", asctime(loctime));
	fprintf(LogFile, "Model name : %s\n", TunableParameters[idx].modelName);
	fprintf(LogFile, "Block name : %s\n", TunableParameters[idx].blockName);
	fprintf(LogFile, "Param name : %s\n", TunableParameters[idx].paramName);
}

int ParametersPanel::UpdateParameter(int idx, double val, struct tm *loctime)
{
	switch (TunableParameters[idx].dataType) {
		case DOUBLE_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %G\n", TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %G\n\n", val);
			}
			TunableParameters[idx].dataValue = val;
                        return 1;
		case FLOAT_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %G\n", (float)TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %G\n\n", (float)val);
			}
			TunableParameters[idx].dataValue = (float)val;
			return 1;
		case INT8_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %d\n", (signed char)TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %d\n\n", (signed char)val);
			}
			TunableParameters[idx].dataValue = (signed char)val;
			return 1;
		case UINT8_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %d\n", (unsigned char)TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %d\n\n", (unsigned char)val);
			}
			TunableParameters[idx].dataValue = (unsigned char)val;
			return 1;
		case INT16_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %d\n", (signed short int)TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %d\n\n", (signed short int)val);
			}
			TunableParameters[idx].dataValue = (signed short int)val;
			return 1;
		case UINT16_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %d\n", (unsigned short int)TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %d\n\n", (unsigned short int)val);
			}
			TunableParameters[idx].dataValue = (unsigned short int)val;
			return 1;
		case INT32_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %d\n", (signed int)TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %d\n\n", (signed int)val);
			}
			TunableParameters[idx].dataValue = (signed int)val;
			return 1;
		case UINT32_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %d\n", (unsigned int)TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %d\n\n", (unsigned int)val);
			}
			TunableParameters[idx].dataValue = (unsigned int)val;
			return 1;
		case BOOLEAN_T:
			if (LogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(LogFile, " Old value : %d\n", (unsigned char)TunableParameters[idx].dataValue);
				fprintf(LogFile, " New value : %d\n\n", (unsigned char)val);
			}
			TunableParameters[idx].dataValue = (unsigned char)val;
			return 1;
	}
	return 0;
}

void ParametersPanel::GetParameters()
{
	for (int i = 0; i < NumTunableParameters; i++) {
		char scalar_value[15];
		sprintf(scalar_value, "%G", TunableParameters[i].dataValue);
		Entry[i]->value(strdup(scalar_value));
	}
}

inline void ParametersPanel::cb_batchdownload_i(Fl_Check_Button *o, void *v)
{
	if (BatchDownload->value()) {
		Download->activate();
	} else {
		Download->deactivate();
	}
	counter = 0;
}

void ParametersPanel::cb_batchdownload(Fl_Check_Button *o, void *v)
{
	((ParametersPanel *)(o->parent()->parent()->parent()->user_data()))->cb_batchdownload_i(o,v);
}

inline void ParametersPanel::cb_log_parameters_i(Fl_Check_Button *o, void *v)
{
	if (o->value()) {
		if ((LogFile = fopen(LogFileName, "a+")) == NULL) {
			fl_alert("Error in opening file %s", LogFileName);
			o->value(0);
			return;
		}
		LogFlag = true;
	} else {
		if (LogFile) {
			fclose(LogFile);
		}
		LogFlag = false;
	}
}

void ParametersPanel::cb_log_parameters(Fl_Check_Button *o, void *v)
{
	((ParametersPanel *)(o->parent()->parent()->parent()->user_data()))->cb_log_parameters_i(o,v);
}

inline void ParametersPanel::cb_close_i(Fl_Button *o, void *v)
{
	Window->hide();
	CommandManager->activate_ChangeParams();
}

void ParametersPanel::cb_close(Fl_Button *o, void *v)
{
	((ParametersPanel *)(o->parent()->parent()->parent()->user_data()))->cb_close_i(o,v);
}

ParametersPanel::ParametersPanel()
{
	Fl_Button *block;
	Fl_Window *w;
	{ Fl_Window *o = Window = new Fl_Window(350, 550);
	  w = o;
	  o->label("Parameters Panel");
	  o->user_data((void *)(this));
	  { Fl_Group *o = new Fl_Group(0, 0, 350, 550);
	    { Fl_Group *o = ParamsInfo = new Fl_Group(0, 0, 350, 550);
	      { Fl_Button *o = new Fl_Button(5, 10, 90, 35, "Model name: ");
		o->labelsize(12);
		o->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
		o->box(FL_NO_BOX);
	      }
	      { Fl_Button *o = new Fl_Button(100, 10, 90, 35, strdup(TunableParameters[0].modelName));
		o->labelsize(14);
		o->labeltype(FL_EMBOSSED_LABEL);
		o->labelfont(3);
		o->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
		o->box(FL_NO_BOX);
	      }
	      { Fl_Scroll *o = ParamsScroll = new Fl_Scroll(5, 60, 345, 290);
		o->type(Fl_Scroll::VERTICAL_ALWAYS);
		o->box(FL_EMBOSSED_BOX);
		Entry = new Fl_Input*[NumTunableParameters];
		for (int i = 0; i < NumTunableParameters; i++) {
			Fl_Box *box = new Fl_Box(FL_FRAME_BOX, 8, 60+35*i, 325, 35, 0);
			box->color(FL_GRAY);
			if (i > 0) {
				if (strcmp(TunableParameters[i-1].blockName, TunableParameters[i].blockName)) {
					block = new Fl_Button(8, 60+35*i, 110, 35, strdup(TunableParameters[i].blockName+strlen(TunableParameters[0].modelName)+1));
				} else {
					block = new Fl_Button(8, 60+35*i, 110, 35, " ");
				}
			} else {
				block = new Fl_Button(8, 60+35*i, 110, 35, strdup(TunableParameters[i].blockName+strlen(TunableParameters[0].modelName)+1));
			}
			block->labelsize(10);
			block->labelfont(9);
			block->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
			block->box(FL_NO_BOX);
			{ Fl_Button *param = new Fl_Button(120, 60+35*i, 110, 35, strdup(TunableParameters[i].paramName));
			  param->labelsize(10);
			  param->align(FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
			  param->box(FL_NO_BOX);
			}
			char scalar_value[15];
			{ Fl_Input *o = Entry[i] = new Fl_Input(250, 65+35*i, 70, 25);
			  sprintf(scalar_value, "%G", TunableParameters[i].dataValue);
			  o->value(strdup(scalar_value));
			  o->when(FL_WHEN_ENTER_KEY);
			  o->callback((Fl_Callback *)cb_update, (void *)i);
			}
		}
		ParamsScroll->end();
	      }
	      { Fl_Check_Button *o = BatchDownload = new Fl_Check_Button(10, 360, 110, 25, "Batch download");
		o->labelsize(12);
		o->selection_color(FL_BLACK);
		o->callback((Fl_Callback *)cb_batchdownload, 0);
	      }
	      { Fl_Button *o = Download = new Fl_Button(10, 385, 110, 30, "Download");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_batch_update, 0);
		o->deactivate();
	      }
	      { Fl_Button *o = GetP = new Fl_Button(130, 385, 110, 30, "Update");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_get_parameters, 0);
	      }
	      LogFileName = strdup("ParamLog");
	      { Fl_Check_Button *o = LogP = new Fl_Check_Button(250, 385, 110, 30, "Logging");
		o->labelsize(12);
		o->selection_color(FL_BLACK);
		o->callback((Fl_Callback *)cb_log_parameters, 0);
	      }
	      { Fl_Button *o = new Fl_Button(278, 520, 70, 25, "Close");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_close);
	      }
	      o->end();
	    }
	    o->end();
          }
	}
	LogFlag = false;
};
