/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termio.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <math.h>

#include <arpa/inet.h>

#include <sys/mman.h>
#include <sys/poll.h>

#define KEEP_STATIC_INLINE
#include <rtai_lxrt_user.h>
#include <net_rpc.h>
#include "rtailab.h"

#define PAR_LOG_FILE_NAME  "PARAMLOG.dat"

static char TargetInterfaceTaskName[7] = "IFTASK";
static char TargetMbxID[4]             = "RTS";
static char TargetLogMbxID[4]          = "RTL";
static char TargetIP[16]               = "127.0.0.1";

static char ParamLogFilePath[MAX_NAME_SIZE] = "./";
static FILE *ParamLogFile;
static volatile int ParamLogFlag;

static char SignalLogFilePath[MAX_NAME_SIZE] = "./";
static FILE *SignalLogFile[MAX_RTAI_SCOPES];
static volatile int SignalLogFlag[MAX_RTAI_SCOPES];
static int ScopeToScreen;

static char LogFilePath[MAX_NAME_SIZE] = "./";
static FILE *LogLogFile[MAX_RTAI_LOGS];
static volatile int LogFlag[MAX_RTAI_LOGS];
static int LogToScreen;

static volatile int endTarget;

static int nTunableParams, nRTAIScopes, nRTAILogs;
static rtTargetParamInfo *TunableParameters;
static rtTargetScopeInfo ScopesVector[MAX_RTAI_SCOPES];
static rtTargetLogInfo LogsVector[MAX_RTAI_LOGS];

static int BatchDownloadCnt;
static rtBatchDownload BatchParams[MAX_BATCH_PARAMS];

static float TargetTime;
static unsigned long long llTargetTime;

static pthread_t rt_TargetInterfaceThread;
static pthread_t rt_GetScopeDataThread[NSCOPETHREADS];
static pthread_t rt_GetLogDataThread[NLOGTHREADS];
static pthread_t rt_GetTargetTimeThread;

static RT_TASK *rt_TargetInterfaceTask;

static long TargetNode;
static int isTargetConnected;
static int isTargetRunning;
static int BatchDownload;

#define msleep(t)  do { poll(0, 0, t); } while (0)

static unsigned long get_an_id(const char *root)
{
        int i;
	char name[7];
        for (i = 0; i < MAX_NHOSTS; i++) {
                sprintf(name, "%s%d", root, i);
                if (!rt_get_adr(nam2num(name))) {
                	return nam2num(name);
                }
        }
        return 0;
}

static char get_key(void)
{
	static struct termio my_kb, orig_kb;
	static int first = 1;
	char ch;
	if (first) {
		first = 0;
		ioctl(0, TCGETA, &orig_kb);
		my_kb = orig_kb;
		my_kb.c_lflag &= ~(ECHO | ICANON);
		my_kb.c_cc[4] = 1;
	}
	ioctl(0, TCSETAF, &my_kb);
	ch = getchar();
	ioctl(0, TCSETAF, &orig_kb);
	return ch;
}

static void *rt_GetTargetTime(void *arg)
{
	RT_TASK *rt_GetTargetTimeTask;

	rt_allow_nonroot_hrt();
	if (!(rt_GetTargetTimeTask = rt_task_init_schmod(get_an_id("HGT"), rt_GetTargetTimePriority, 0, 0, SCHED_FIFO, 0xFF))) {
		printf("Cannot init rt_GetTargetTimeTask\n");
		return (void *)1;
	}
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (isTargetConnected) {
		rt_send(rt_TargetInterfaceTask, GET_TARGET_TIME);
		msleep(100);
	}
	rt_task_delete(rt_GetTargetTimeTask);
	return 0;
}

static void *rt_GetLogData(void *arg)
{
	int index = ((int *)arg)[0];
	char name[7];
	FILE *LogFile;
	int i, k, j, n, nData, max_msg_len, msg_len, max_log_bytes, DataCnt;
	RT_TASK *rt_GetLogDataTask;
	MBX *rt_GetLogDataMbx;
	long rt_GetLogDataPort;
	float MsgBuf[MAX_MSG_LEN/sizeof(float)];

	rt_allow_nonroot_hrt();
	if (!(rt_GetLogDataTask = rt_task_init_schmod(get_an_id("HGL"), rt_GetLogPriority, 0, 0, SCHED_RR, 0xFF))) {
		printf("Cannot init Host Get and Log DataTask\n");
		return (void *)1;
	}
	rt_GetLogDataPort = rt_request_port(TargetNode);
	sprintf(name, "%s%d", TargetLogMbxID, index);
	if (!(rt_GetLogDataMbx = (MBX *)RT_get_adr(TargetNode, rt_GetLogDataPort, name))) {
		printf("Error in getting %s mailbox address\n", name);
		return (void *)1;
	}
	max_log_bytes = LogsVector[index].nrow*LogsVector[index].ncol*sizeof(float);
	max_msg_len = (MAX_MSG_LEN/max_log_bytes);
	msg_len = (((int)(max_log_bytes*REFRESH_RATE/LogsVector[index].dt))/max_log_bytes + 1);
	if (msg_len > max_msg_len) {
		msg_len = max_msg_len;
	}
	msg_len *= max_log_bytes;
	nData = msg_len/max_log_bytes;
	DataCnt = 0;
	rt_send(rt_TargetInterfaceTask, 0);
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (1) {
		while (RT_mbx_receive_if(TargetNode, rt_GetLogDataPort, rt_GetLogDataMbx, &MsgBuf, msg_len)) {
			if (!isTargetConnected) goto end;
			msleep(10);
		}
		if (LogFlag[index]) {
			LogFile = LogToScreen ? stdout : LogLogFile[index];
			for (n = 0; n < nData; n++) {
				fprintf(LogFile, "Data # %d\n", ++DataCnt);
				for (i = 0; i < LogsVector[index].nrow; i++) {
					j = n*LogsVector[index].nrow*LogsVector[index].ncol + i;
					for (k = 0; k < LogsVector[index].ncol; k++) {
						fprintf(LogFile, "%1.5f ", MsgBuf[j]);
						j += LogsVector[index].nrow;
					}
					fprintf(LogFile, "\n");
				}
			}
		}
	}
end:	rt_release_port(TargetNode, rt_GetLogDataPort);
	rt_task_delete(rt_GetLogDataTask);
	return 0;
}

static void *rt_GetScopeData(void *arg)
{
	int index = ((int *)arg)[0];
	char name[7];
	FILE *ScopeFile;
	int i, k, j, nData, max_msg_len, msg_len, max_traces_bytes;
	RT_TASK *rt_GetScopeDataTask;
	MBX *rt_GetScopeDataMbx;
	long rt_GetScopeDataPort;
	float MsgBuf[MAX_MSG_LEN/sizeof(float)];

	rt_allow_nonroot_hrt();
	if (!(rt_GetScopeDataTask = rt_task_init_schmod(get_an_id("HGS"), rt_GetScopePriority, 0, 0, SCHED_RR, 0xFF))) {
		printf("Cannot init Host Collect and Scope DataTask\n");
		return (void *)1;
	}
	rt_GetScopeDataPort = rt_request_port(TargetNode);
	sprintf(name, "%s%d", TargetMbxID, index);
	if (!(rt_GetScopeDataMbx = (MBX *)RT_get_adr(TargetNode, rt_GetScopeDataPort, name))) {
		printf("Error in getting %s mailbox address\n", name);
		return (void *)1;
	}
	max_traces_bytes = (ScopesVector[index].ntraces + 1)*sizeof(float);
//	max_traces_bytes = ScopesVector[index].ntraces*sizeof(float);
	max_msg_len = (MAX_MSG_LEN/max_traces_bytes);
	msg_len = (((int)(max_traces_bytes*REFRESH_RATE/ScopesVector[index].dt))/max_traces_bytes + 1);
	if (msg_len > max_msg_len) {
		msg_len = max_msg_len;
	}
	msg_len *= max_traces_bytes;
	nData = msg_len/max_traces_bytes;
	rt_send(rt_TargetInterfaceTask, 0);
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (1) {
		while (RT_mbx_receive_if(TargetNode, rt_GetScopeDataPort, rt_GetScopeDataMbx, &MsgBuf, msg_len)) {
			if (!isTargetConnected) goto end;
			msleep(10);
		}
		if (SignalLogFlag[index]) {
			ScopeFile = ScopeToScreen ? stdout : SignalLogFile[index];
			j = 0;
			for (i = 0; i < nData; i++) {
//				fprintf(ScopeFile, "%llu ", ++llTargetTime);
				for (k = 0; k < ScopesVector[index].ntraces + 1; k++) {
					fprintf(ScopeFile, "%1.5f ", MsgBuf[j++]);
				}
				fprintf(ScopeFile, "\n");
			}
		}
	}
end:	rt_release_port(TargetNode, rt_GetScopeDataPort);
	rt_task_delete(rt_GetScopeDataTask);
	return 0;
}

static void SignalLogHandler(int n);
static void LogHandler(int n);
static void ParamLogHandler(void);

static void end_support_threads(void)
{
	int n;
	isTargetConnected = 0;
	for (n = 0; n < nRTAIScopes; n++) {
		if (SignalLogFlag[n]) {
			SignalLogHandler(n);
		}
		pthread_join(rt_GetScopeDataThread[n], NULL);
	}
	for (n = 0; n < nRTAILogs; n++) {
		if (LogFlag[n]) {
			LogHandler(n);
		}
		pthread_join(rt_GetLogDataThread[n], NULL);
	}
	pthread_join(rt_GetTargetTimeThread, NULL);
	if (ParamLogFlag) {
		ParamLogHandler();
	}
}

static void *rt_TargetInterface(void *args)
{
	unsigned int msg;
	int code, n;
	long TargetPort;
	RT_TASK *IfTask, *task;

	rt_allow_nonroot_hrt();
	if (!(rt_TargetInterfaceTask = rt_task_init_schmod(get_an_id("HTI"), rt_TargetInterfacePriority, 0, 0, SCHED_FIFO, 0xFF))) {
		printf("Cannot init rt_TargetInterfaceTask\n");
		return (void *)1;
	}
	TargetPort = 0;
	IfTask = 0;
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (!endTarget) {
		if (!(task = rt_receive(0, &code))) continue;
		switch (code & 0xFFFF) {
			case CONNECT_TO_TARGET: {
        			struct sockaddr_in TargetAddr;
				rtTargetParamInfo TargetParameters;
				int ConnectionPollingCounter;
				char name[7];

				if (TargetIP[0] == '0') {
					TargetNode = 0;
				} else {
					printf("Trying to connect to %s ...\n", TargetIP);
					inet_aton(TargetIP, &TargetAddr.sin_addr);
					TargetNode = TargetAddr.sin_addr.s_addr;
					ConnectionPollingCounter = MAX_CONNECT_TRY;
					while ((TargetPort = rt_request_port(TargetNode)) <= 0 && TargetPort != -EINVAL && !ConnectionPollingCounter--) {
						msleep(100);
					}
					if (ConnectionPollingCounter <= 0) {
						printf("abort, no node.\n");
						rt_return(task, CONNECT_TO_TARGET);
						break;
					}
				}
				if (!(IfTask = (RT_TASK *)RT_get_adr(TargetNode, TargetPort, TargetInterfaceTaskName))) {
					printf("abort, no target task.\n");
					rt_return(task, CONNECT_TO_TARGET);
					break;
				}
				printf("... connected to target node: %s, at port %ld\n", TargetIP, TargetPort);
/* Receiving running state and number of tunable params from remote target */
				RT_rpc(TargetNode, TargetPort, IfTask, 'c', &isTargetRunning);
				nTunableParams = isTargetRunning & 0xffff;
				isTargetRunning >>= 16;
				TunableParameters = (rtTargetParamInfo *)malloc(nTunableParams*sizeof(rtTargetParamInfo));
/* Receiving tunable paramaters info from remote target */
				for (n = 0; n < nTunableParams; n++) {
					RT_rpcx(TargetNode, TargetPort, IfTask, "i", &TargetParameters, sizeof(char), sizeof(TargetParameters));
					strncpyz(TunableParameters[n].modelName, TargetParameters.modelName, MAX_NAME_SIZE);
					strncpyz(TunableParameters[n].blockName, TargetParameters.blockName, MAX_NAME_SIZE);
					strncpyz(TunableParameters[n].paramName, TargetParameters.paramName, MAX_NAME_SIZE);
					TunableParameters[n].dataType  = TargetParameters.dataType;
					TunableParameters[n].dataClass = TargetParameters.dataClass;
					TunableParameters[n].dataValue = TargetParameters.dataValue;
					strncpyz(TunableParameters[n].modelName, TargetParameters.modelName, MAX_NAME_SIZE);
				}
				for (n = 0; n < MAX_RTAI_SCOPES; n++) {
					char ScopeName[MAX_NAME_SIZE];
					sprintf(name, "%s%d", TargetMbxID, n);
					if (!RT_get_adr(TargetNode, TargetPort, name)) {
						break;
					}
					RT_rpcx(TargetNode, TargetPort, IfTask, &n, &ScopesVector[n].ntraces, sizeof(int), sizeof(int));
					printf("Scope %s, # of traces %d.\n", name, ScopesVector[n].ntraces);
					RT_rpcx(TargetNode, TargetPort, IfTask, &n, &ScopeName, sizeof(int), sizeof(ScopeName));
					strncpyz(ScopesVector[n].name, ScopeName, MAX_NAME_SIZE);
					RT_rpcx(TargetNode, TargetPort, IfTask, &n, &ScopesVector[n].dt, sizeof(int), sizeof(float));
				}
				nRTAIScopes = n;
/* Inform remote target that all params infos have been acquired */
				n = -1;
				RT_rpcx(TargetNode, TargetPort, IfTask, &n, name, sizeof(int), sizeof(name));
				for (n = 0; n < MAX_RTAI_LOGS; n++) {
					char LogName[MAX_NAME_SIZE];
					sprintf(name, "%s%d", TargetLogMbxID, n);
					if (!RT_get_adr(TargetNode, TargetPort, name)) {
						break;
					}
					RT_rpcx(TargetNode, TargetPort, IfTask, &n, &LogsVector[n].nrow, sizeof(int), sizeof(int));
					RT_rpcx(TargetNode, TargetPort, IfTask, &n, &LogsVector[n].ncol, sizeof(int), sizeof(int));
					RT_rpcx(TargetNode, TargetPort, IfTask, &n, &LogName, sizeof(int), sizeof(LogName));
					strncpyz(LogsVector[n].name, LogName, MAX_NAME_SIZE);
					RT_rpcx(TargetNode, TargetPort, IfTask, &n, &LogsVector[n].dt, sizeof(int), sizeof(float));
				}
				nRTAILogs = n;
/* Inform remote target that all params infos have been acquired */
				n = -1;
				RT_rpcx(TargetNode, TargetPort, IfTask, &n, name, sizeof(int), sizeof(name));
				isTargetConnected = 1;
/* Create ScopeData threads */
				for (n = 0; n < nRTAIScopes; n++) {
					pthread_create(&rt_GetScopeDataThread[n], NULL, rt_GetScopeData, &n);
					rt_receive(0, &msg);
				}
/* Create LogData threads */
				for (n = 0; n < nRTAILogs; n++) {
					pthread_create(&rt_GetLogDataThread[n], NULL, rt_GetLogData, &n);
					rt_receive(0, &msg);
				}
				printf("Connected to %s\n", TunableParameters[0].modelName);
/* Create GetTargetTime thread */
				pthread_create(&rt_GetTargetTimeThread, NULL, rt_GetTargetTime, NULL);
				rt_return(task, CONNECT_TO_TARGET);
				break;
			}
			case DISCONNECT_FROM_TARGET: {
				end_support_threads();
				rt_release_port(TargetNode, TargetPort);
				TargetPort = TargetNode = 0;
				free(TunableParameters);
				TunableParameters = 0;
				rt_return(task, DISCONNECT_FROM_TARGET);
				break;
			}
			case START_TARGET: {
				if (!isTargetRunning) {
					RT_rpc(TargetNode, TargetPort, IfTask, 's', &isTargetRunning);
				}
				rt_return(task, START_TARGET);
				break;
			}
			case STOP_TARGET: {
				if (isTargetRunning) {
					end_support_threads();
					RT_rpc(TargetNode, TargetPort, IfTask, 't', &isTargetRunning);
					rt_release_port(TargetNode, TargetPort);
					TargetNode = TargetPort = 0;
					free(TunableParameters);
					TunableParameters = 0;
				}
				rt_return(task, STOP_TARGET);
				break;
			}
			case UPDATE_PARAM: {
				n = code >> 16;
				printf("%f\n", TunableParameters[n].dataValue);
				RT_rpc(TargetNode, TargetPort, IfTask, 'p', &isTargetRunning);
				RT_rpcx(TargetNode, TargetPort, IfTask, &n, &msg, sizeof(int), sizeof(int));
				RT_rpcx(TargetNode, TargetPort, IfTask, &TunableParameters[n].dataValue, &msg, sizeof(double), sizeof(int));
				rt_return(task, UPDATE_PARAM);
				break;
			}
			case BATCH_DOWNLOAD: {
				RT_rpc(TargetNode, TargetPort, IfTask, 'd', &isTargetRunning);
				RT_rpcx(TargetNode, TargetPort, IfTask, &BatchDownloadCnt, &msg, sizeof(int), sizeof(int));
				RT_rpcx(TargetNode, TargetPort, IfTask, &BatchParams, &msg, sizeof(rtBatchDownload)*BatchDownloadCnt, sizeof(int));
				for (n = 0; n < BatchDownloadCnt; n++) {
					printf("%f\n", BatchParams[n].value);
				}
				rt_return(task, BATCH_DOWNLOAD);
				break;
			}
			case GET_TARGET_TIME: {
				RT_rpc(TargetNode, TargetPort, IfTask, 'm', &isTargetRunning);
				RT_rpcx(TargetNode, TargetPort, IfTask, &msg, &TargetTime, sizeof(int), sizeof(float));
				llTargetTime = TargetTime*1.0E6;
				break;
			}
			case GET_PARAMS: {
				RT_rpc(TargetNode, TargetPort, IfTask, 'g', &isTargetRunning);
				for (n = 0; n < nTunableParams; n++) {
					RT_rpcx(TargetNode, TargetPort, IfTask, "g", &TunableParameters[n].dataValue, 1, sizeof(double));
				}
				rt_return(task, GET_PARAMS);
				break;
			}
		}
	}
	rt_task_delete(rt_TargetInterfaceTask);
	return 0;
}

static void ConnectDisconnect(void)
{
	unsigned int rep;
	if (!isTargetConnected) {
		rt_rpc(rt_TargetInterfaceTask, CONNECT_TO_TARGET, &rep);
	} else {
		rt_rpc(rt_TargetInterfaceTask, DISCONNECT_FROM_TARGET, &rep);
	}
}

static void StartStop(void)
{
	unsigned int rep;
	if (isTargetConnected) {
		if (!isTargetRunning) {
			rt_rpc(rt_TargetInterfaceTask, START_TARGET, &rep);
		} else {
			rt_rpc(rt_TargetInterfaceTask, STOP_TARGET, &rep);
		}
	}
}

static void LogTimeAndNames(int idx, struct tm *loctime)
{
	fprintf(ParamLogFile, "Time       : %s", asctime(loctime));
	fprintf(ParamLogFile, "Model name : %s\n", TunableParameters[idx].modelName);
	fprintf(ParamLogFile, "Block name : %s\n", TunableParameters[idx].blockName);
	fprintf(ParamLogFile, "Param name : %s\n", TunableParameters[idx].paramName);
}

static int do_UpdateParamValue(int idx, double val, struct tm *loctime)
{
	switch (TunableParameters[idx].dataType) {
		case DOUBLE_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %G\n", TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %G\n\n", val);
			}
			TunableParameters[idx].dataValue = val;
			return 1;
		}
		case FLOAT_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %G\n", (float)TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %G\n\n", (float)val);
			}
			TunableParameters[idx].dataValue = (float)val;
			return 1;
		}
		case INT8_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %d\n", (signed char)TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %d\n\n", (signed char)val);
			}
			TunableParameters[idx].dataValue = (signed char)val;
			return 1;
		}
		case UINT8_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %d\n", (unsigned char)TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %d\n\n", (unsigned char)val);
			}
			TunableParameters[idx].dataValue = (unsigned char)val;
			return 1;
		}
		case INT16_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %d\n", (signed short int)TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %d\n\n", (signed short int)val);
			}
			TunableParameters[idx].dataValue = (signed short int)val;
			return 1;
		}
		case UINT16_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %d\n", (unsigned short int)(unsigned short int)TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %d\n\n", (unsigned short int)val);
			}
			TunableParameters[idx].dataValue = (unsigned short int)val;
			return 1;
		}
		case INT32_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %d\n", (signed int)TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %d\n\n", (signed int)val);
			}
			TunableParameters[idx].dataValue = (signed int)val;
			return 1;
		}
		case UINT32_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %d\n", (unsigned int)TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %d\n\n", (unsigned int)val);
			}
			TunableParameters[idx].dataValue = (unsigned int)val;
			return 1;
		}
		case BOOLEAN_T: {
			if (ParamLogFlag) {
				LogTimeAndNames(idx, loctime);
				fprintf(ParamLogFile, " Old value : %d\n", (unsigned char)TunableParameters[idx].dataValue);
				fprintf(ParamLogFile, " New value : %d\n\n", (unsigned char)val);
			}
			TunableParameters[idx].dataValue = (unsigned char)val;
			return 1;
		}
	}
	return 0;
}

static void UpdateParamValue(int idx, double val)
{
	time_t curtime = time(NULL);
	struct tm *loctime = localtime(&curtime);

	if (!BatchDownload) {
		if (do_UpdateParamValue(idx, val, loctime)) {
			unsigned int rep;
			rt_rpc(rt_TargetInterfaceTask, (idx << 16) | UPDATE_PARAM, &rep);
		}
	} else if (BatchDownloadCnt < MAX_BATCH_PARAMS) {
		BatchParams[BatchDownloadCnt].index = idx;
		BatchParams[BatchDownloadCnt++].value = val;
	} else {
		printf("Batch download buffer is full (%d), you should dowload to make room\n<enter anything to continue>\n", MAX_BATCH_PARAMS);
		get_key();
	}
}


static void BatchUpdateParamValue(void)
{
	int i, n;
	time_t curtime = time(NULL);
	struct tm *loctime = localtime(&curtime);

	for (i = n = 0; i < BatchDownloadCnt; i++) {
		n += do_UpdateParamValue(BatchParams[i].index, BatchParams[i].value, loctime);
	}
	if (BatchDownloadCnt > 0 && n == BatchDownloadCnt) {
		unsigned int rep;
		rt_rpc(rt_TargetInterfaceTask, BATCH_DOWNLOAD, &rep);
	}
}

static void ParamLogHandler(void)
{
	time_t curtime;
	struct tm *loctime;
	int i;

	if (!ParamLogFlag) {
		char filename[MAX_NAME_SIZE];
		strncpyz(filename, ParamLogFilePath, MAX_NAME_SIZE);
		strncat(filename, PAR_LOG_FILE_NAME, MAX_NAME_SIZE);
		if ((ParamLogFile = fopen(filename, "a+")) == NULL) {
			fprintf(stderr, "Error in opening file %s", PAR_LOG_FILE_NAME);
			fprintf(stderr, "Bad file name!");
			ParamLogFlag = 0;
			return;
		}
		curtime = time(NULL);
	  	loctime = localtime(&curtime);
		fprintf(ParamLogFile, "Target Parameters Status\n");
		fprintf(ParamLogFile, "========================\n\n");
		fprintf(ParamLogFile, "%s\n", asctime(loctime));
		for (i = 0; i < nTunableParams; i++) {
			fprintf(ParamLogFile, "Block name : %s\n", TunableParameters[i].blockName);
			fprintf(ParamLogFile, "Param name : %s\n", TunableParameters[i].paramName);
			switch (TunableParameters[i].dataType) {
				case DOUBLE_T:
					fprintf(ParamLogFile, "Data type  : DOUBLE\n");
					fprintf(ParamLogFile, "Data value : %G\n", TunableParameters[i].dataValue);
					break;
				case FLOAT_T:
					fprintf(ParamLogFile, "Data type  : FLOAT\n");
					fprintf(ParamLogFile, "Data value : %G\n", TunableParameters[i].dataValue);
					break;
				case INT8_T:
					fprintf(ParamLogFile, "Data type  : INT 8\n");
					fprintf(ParamLogFile, "Data value : %d\n", (signed char)TunableParameters[i].dataValue);
					break;
				case UINT8_T:
					fprintf(ParamLogFile, "Data type  : UNSIGNED INT 8\n");
					fprintf(ParamLogFile, "Data value : %d\n", (unsigned char)TunableParameters[i].dataValue);
					break;
				case INT16_T:
					fprintf(ParamLogFile, "Data type  : INT 16\n");
					fprintf(ParamLogFile, "Data value : %d\n", (signed short int)TunableParameters[i].dataValue);
					break;
				case UINT16_T:
					fprintf(ParamLogFile, "Data type  : UNSIGNED INT 16\n");
					fprintf(ParamLogFile, "Data value : %d\n", (unsigned short int)TunableParameters[i].dataValue);
					break;
				case INT32_T:
					fprintf(ParamLogFile, "Data type  : INT 32\n");
					fprintf(ParamLogFile, "Data value : %d\n", (signed int)TunableParameters[i].dataValue);
					break;
				case UINT32_T:
					fprintf(ParamLogFile, "Data type  : UNSIGNED INT 32\n");
					fprintf(ParamLogFile, "Data value : %d\n", (unsigned int)TunableParameters[i].dataValue);
					break;
				case BOOLEAN_T:
					fprintf(ParamLogFile, "Data type  : BOOLEAN\n");
					fprintf(ParamLogFile, "Data value : %d\n", (unsigned char)TunableParameters[i].dataValue);
					break;
				default:
					break;
			}
			fprintf(ParamLogFile, "\n");
			ParamLogFlag = 1;
                }
	} else {
                if (ParamLogFile) {
                        fclose(ParamLogFile);
                }
                ParamLogFlag = 0;
        }
}

static void rtChangeParametersManager(void)
{
	int i;
	char str[MAX_NAME_SIZE];

	while (isTargetConnected) {
		printf("\n  MODEL NAME <> %s <>\n", TunableParameters[0].modelName);
		printf("- Get params\n");
		printf("- Change parameters\n");
		printf("- Batch dowload (status: %s)\n", BatchDownload ? "ENABLED" : "DISABLED");
		printf("- Download\n");
		printf("- parameter Logging (status: %s)\n", ParamLogFlag ? "ENABLED" : "DISABLED");
		printf("- param log File path (%s)\n", ParamLogFilePath);
		printf("<upper case for the rest, anything else to end>\n");
		str[0] = get_key();
		switch (tolower(str[0])) {
			case 'f': {
				printf("New param log file path (%s): ", ParamLogFilePath);
				fgets(str, sizeof(TargetIP), stdin);
				str[strlen(str) - 1] = 0;
				if (str[0]) {
					strcpy(ParamLogFilePath, str);
				} 
				break;
			} 
			case 'g': {
				unsigned int rep;
				rt_rpc(rt_TargetInterfaceTask, GET_PARAMS, &rep);
				break;
			} 
			case 'l': {
				ParamLogHandler();
				break;
			} 
			case 'b': {
				BatchDownload = 1 - BatchDownload;
				BatchDownloadCnt = 0;
				break;
			} 
			case 'd': {
				BatchUpdateParamValue();
				BatchDownloadCnt = 0;
				break;
			} 
			case 'c': {
				while (1) {
					printf("\n");
					for (i = 0; i < nTunableParams; i++) {
						printf("- (%d) %s - %s: %f\n", i, TunableParameters[i].blockName, TunableParameters[i].paramName, TunableParameters[i].dataValue);
					}
					printf("<index for variables choice, anything else to end>\n");
					fgets(str, MAX_NAME_SIZE, stdin);
					if (isdigit(str[0])) {
						i = atoi(str);
						printf("New %s - %s (%d): %f < ", TunableParameters[i].blockName, TunableParameters[i].blockName, i, TunableParameters[i].dataValue);
						fgets(str, MAX_NAME_SIZE, stdin);
						if (isspace(str[0])) {
							continue;
						}
						UpdateParamValue(i, atof(str));
					} else {
						break;
					}
				} 
				break;
			} 
			default: return;
		} 
	}
}

static void LogHandler(int n)
{
	if (!LogFlag[n]) {
		char filename[MAX_NAME_SIZE];
		strncpyz(filename, LogFilePath, MAX_NAME_SIZE);
		strncat(filename, LogsVector[n].name, MAX_NAME_SIZE);
		strncat(filename, ".dat", 4);
		if ((LogLogFile[n] = fopen(filename, "a+")) == NULL) {
			printf("Error in opening file %s", filename);
			return;
		}
		LogFlag[n] = 1;
	} else {
		if (LogLogFile[n]) {
			fclose(LogLogFile[n]);
		}
		LogFlag[n] = 0;
	}
}

static void SignalLogHandler(int n)
{
	if (!SignalLogFlag[n]) {
		char filename[MAX_NAME_SIZE];
		strncpyz(filename, SignalLogFilePath, MAX_NAME_SIZE);
		strncat(filename, ScopesVector[n].name, MAX_NAME_SIZE);
		strncat(filename, ".dat", 4);
		if ((SignalLogFile[n] = fopen(filename, "a+")) == NULL) {
			printf("Error in opening file %s", filename);
			return;
		}
		SignalLogFlag[n] = 1;
	} else {
		if (SignalLogFile[n]) {
			fclose(SignalLogFile[n]);
		}
		SignalLogFlag[n] = 0;
	}
}

static void rtLogManager(void)
{
	int i;
	char str[MAX_NAME_SIZE];

	while (isTargetConnected) {
		printf("\n  MODEL NAME <> %s <>\n", TunableParameters[0].modelName);
		for (i = 0; i < nRTAILogs; i++) {
			printf("- (%d) %s (status: %s)\n", i, LogsVector[i].name, LogFlag[i] ? "ENABLED" : "DISABLED");
		}
		printf("- divert to Screen (status: %s)\n", LogToScreen ? "ENABLED" : "DISABLED");
		printf("- log file Path (%s)\n", LogFilePath);
		printf("<index for logs choice, upper case for the rest, anything else to end>\n");
		str[0] = get_key();
		if (tolower(str[0]) == 'p') {
			printf("New log file path (%s): ", LogFilePath);
			fgets(str, sizeof(TargetIP), stdin);
			str[strlen(str) - 1] = 0;
			if (str[0]) {
				strcpy(LogFilePath, str);
			} 
		} else if (tolower(str[0]) == 's') {
			LogToScreen = 1 - LogToScreen;
		} else if (nRTAILogs && isdigit(str[0])) {
			LogHandler(atoi(str));
		} else { 
			return;
		} 
	}

}

static void rtScopeManager(void)
{
	int i;
	char str[MAX_NAME_SIZE];

	while (isTargetConnected) {
		printf("\n  MODEL NAME <> %s <>\n", TunableParameters[0].modelName);
		for (i = 0; i < nRTAIScopes; i++) {
			printf("- (%d) %s (status: %s)\n", i, ScopesVector[i].name, SignalLogFlag[i] ? "ENABLED" : "DISABLED");
		}
		printf("- divert to Screen (status: %s)\n", ScopeToScreen ? "ENABLED" : "DISABLED");
		printf("- signal log file Path (%s)\n", SignalLogFilePath);
		printf("<index for scopes choice, upper case for the rest, anything else to end>\n");
		str[0] = get_key();
		if (tolower(str[0]) == 'p') {
			printf("New signal log file path (%s): ", SignalLogFilePath);
			fgets(str, sizeof(TargetIP), stdin);
			str[strlen(str) - 1] = 0;
			if (str[0]) {
				strcpy(SignalLogFilePath, str);
			} 
		} else if (tolower(str[0]) == 's') {
			ScopeToScreen = 1 - ScopeToScreen;
		} else if (nRTAIScopes && isdigit(str[0])) {
			SignalLogHandler(atoi(str));
		} else { 
			return;
		} 
	}

}

static void SelectTarget(void)
{
	char str[MAX_NAME_SIZE];
	while (1) {
		printf("\n  MODEL NAME <> %s <>\n", TunableParameters[0].modelName);
		printf("- target Ip (dotted decimal notation): %s\n", TargetIP);
		printf("- target interface Task name: %s\n", TargetInterfaceTaskName);
		printf("- root of target Scopes mail box names: %s\n", TargetMbxID);
		printf("- root of target Logs mail box names: %s\n", TargetLogMbxID);
		printf("<upper case choice, anything else to end>\n");
		str[0] = get_key();
		switch(tolower(str[0])) {
			case 'i' : {
				printf("new IP: "); 
				fgets(str, sizeof(str), stdin);
				str[strlen(str) - 1] = 0;
				if (str[0]) {
					strncpyz(TargetIP, str, sizeof(TargetIP));
				}
				break;
			} 
			case 't' : {
				printf("new task id: "); 
				fgets(str, sizeof(str), stdin);
				str[strlen(str) - 1] = 0;
				if (str[0]) {
					strncpyz(TargetInterfaceTaskName, str, sizeof(TargetInterfaceTaskName));
				}
				break;
			} 
			case 's' : {
				printf("new scopes mail box id: "); 
				fgets(str, sizeof(str), stdin);
				str[strlen(str) - 1] = 0;
				if (str[0]) {
					strncpyz(TargetMbxID, str, sizeof(TargetMbxID));
				}
				break;
			} 
			case 'l' : {
				printf("new logs mail box id: "); 
				fgets(str, sizeof(str), stdin);
				str[strlen(str) - 1] = 0;
				if (str[0]) {
					strncpyz(TargetLogMbxID, str, sizeof(TargetLogMbxID));
				}
				break;
			} 
			default: return;
		}
	}
}

static void MainMenu(void)
{
	char str[MAX_NAME_SIZE];

	while (1) {
		printf("\n  MODEL NAME <> %s <>\n", TunableParameters[0].modelName);
		printf("- Get target time: %f (s)\n", TargetTime);
		printf("- Remote target setting\n");
		printf("- Connect/Disconnect (status: %s)\n", isTargetConnected ? "CONNECTED" : "NOT CONNECTED");
		printf("- Start/Stop (status: %s)\n", isTargetRunning ? "RUNNING" : "NOT RUNNING");
		printf("- Parameters setting\n");
		printf("- Time histories\n");
		printf("- Log histories\n");
		printf("<upper case choice, anything else to exit RTAI-Lab>\n");
		str[0] = get_key();
		switch(tolower(str[0])) {
			case 'g' : {
				break;
			} 
			case 'r' : {
				SelectTarget();
				break;
			} 
			case 'c' : 
			case 'd' : {
				ConnectDisconnect();
				break;
			} 
			case 's' : {
				StartStop(); 
				break;
			} 
			case 'p' : {
				rtChangeParametersManager();
				break;
			} 
			case 't' : {
				rtScopeManager();
				break;
			} 
			case 'l' : {
				rtLogManager();
				break;
			} 
			default: return;
		}
	}
}

int main(int argc, char **argv)
{
	RT_TASK *MainTask;

	rt_allow_nonroot_hrt();
	if (!(MainTask = rt_task_init_schmod(get_an_id("HMN"), rt_MainPriority, 0, 0, SCHED_FIFO, 0xFF))) {
		printf("Cannot init HostMainTask\n");
		return 1;
	}
	pthread_create(&rt_TargetInterfaceThread, NULL, rt_TargetInterface, NULL);
	MainMenu();
	if (isTargetConnected) {
		int i;
		rt_rpc(rt_TargetInterfaceTask, DISCONNECT_FROM_TARGET, &i);
	}
	endTarget = 1;
	rt_send(rt_TargetInterfaceTask, 0xFFFFFFFF);
	pthread_join(rt_TargetInterfaceThread, NULL);
	rt_task_delete(MainTask);
	return 0;
}
