/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/


#ifndef _RTAILAB_H
#define _RTAILAB_H

#define KEEP_STATIC_INLINE
#include <rtai_lxrt_user.h>
#include <rtai_lxrt.h>
#include <net_rpc.h>

#define MAX_RTAI_SCOPES		1000
#define MAX_RTAI_LOGS		1000
#define MAX_MSG_LEN		(MAX_MSG_SIZE - 100)
#define MAX_NAME_SIZE		50
#define MAX_NHOSTS		1000
#define MAX_TRACE_LENGTH	100000
#define MAX_CHANNELS		5
#define MAX_BATCH_PARAMS	1000
#define MAX_CONNECT_TRY    5

#define REFRESH_RATE		0.05
#define NSCOPETHREADS		100
#define NLOGTHREADS		100

#define rt_GetScopePriority		99
#define rt_GetLogPriority		99
#define rt_TargetInterfacePriority	98
#define rt_GetTargetTimePriority	99
#define rt_MainPriority			99

#define FLTK_EVENTS_TICK        0.01

#define CONNECT_TO_TARGET	0
#define DISCONNECT_FROM_TARGET	1
#define START_TARGET		2
#define STOP_TARGET		3
#define UPDATE_PARAM		4
#define GET_TARGET_TIME		5
#define BATCH_DOWNLOAD		6
#define GET_PARAMS		7
#define CLOSE			8

#define DOUBLE_T        0
#define FLOAT_T         1
#define INT8_T          2
#define UINT8_T         3
#define INT16_T         4
#define UINT16_T        5
#define INT32_T         6
#define UINT32_T        7
#define BOOLEAN_T       8

#define SCALAR_C	0
#define VECTOR_C	1
#define ROW_MATRIX_C	2
#define COL_MATRIX_C	3

#define msleep(t)	do { poll(0, 0, t); } while (0)

static inline void strncpyz(char *dest, const char *src, size_t n)
{
	strncpy(dest, src, n);
	dest[n-1] = '\0';
}

typedef struct rtTargetParamInfoStruct rtTargetParamInfo;
typedef struct rtTargetScopeInfoStruct rtTargetScopeInfo;
typedef struct rtTargetLogInfoStruct rtTargetLogInfo;
typedef struct rtPointStruct rtPoint;
typedef struct rtBatchDownloadStruct rtBatchDownload;

struct rtTargetParamInfoStruct {
	char modelName[MAX_NAME_SIZE];
	char blockName[MAX_NAME_SIZE];
	char paramName[MAX_NAME_SIZE];
	unsigned int dataType;
	unsigned int dataClass;
	double dataValue;
};

struct rtTargetScopeInfoStruct {
	char name[MAX_NAME_SIZE];
	int ntraces;
	float dt;
	int visible;
};

struct rtTargetLogInfoStruct {
	char name[MAX_NAME_SIZE];
	int nrow, ncol;
	float dt;
};

struct rtPointStruct {
	float x;
	float y;
};

struct rtBatchDownloadStruct {
	int index;
	double value;
};

#endif
