/*
COPYRIGHT (C) 2003  Roberto Bucher (roberto.bucher@die.supsi.ch)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include "asm/io.h"

#define BASE_ADDR 0x300

#define RLD 0x00
#define CMR 0x20
#define IOR 0x40
#define IDR 0x60

#define PI	3.14159265358979

#define RESET_DETECT 0x800000
#define INDEX_DETECT 0x000000

static struct{
    int module;
    int res;
    int mode;
    int count;
    int rot;
    int first_index;
    int baseAddr;
}quad4[4];

void cioquad4_init(int port, int module, int res, int mode, int count, int rot)
{
    int baseAddr;

    quad4[port-1].module=module;
    quad4[port-1].res=res;
    quad4[port-1].mode=mode;
    quad4[port-1].count=count;
    quad4[port-1].rot=rot;
    quad4[port-1].first_index=0;

    baseAddr = BASE_ADDR + (module-1)*2;
    quad4[port-1].baseAddr=baseAddr;
    if(mode==4) mode=3;
    outb(CMR | (mode << 3),baseAddr + 0x01);   // Set mode SINGLE, DOUBLE,
                                                           // QUADRUPLE */

    outb(RLD | 0x04,baseAddr + 0x01);          // RESET BT, CT, CPT, S
    outb(RLD | 0x06,baseAddr + 0x01);          // RESET E  

    outb(RLD | 0x01,baseAddr + 0x01);          // RESET BP  

    outb(INDEX_DETECT & 0x0000ff,baseAddr);         // PRESET INDEX_DETECT 
    outb((INDEX_DETECT >> 8) & 0x0000ff,baseAddr);  
    outb((INDEX_DETECT >> 16) & 0x0000ff,baseAddr);

    outb(IOR | 0x00,baseAddr + 0x01);   // DISABLE A/B 
    outb(IDR | 0x03,baseAddr + 0x01);   // ENABLE INDEX POSITIVE

    // Now : Index is enabled, positive and 
    //       Load CNTR Input

    outb(0x0f,BASE_ADDR + 0x08);         // Index TO LCNTR
    outb(0x00,BASE_ADDR + 0x09);         // 4x 24 Bit Counter
    outb(0x00,BASE_ADDR + 0x12);         // DISABLE INTERRUPT
    outb(RLD | 0x08,baseAddr + 0x01);   // PRESET TO COUNTER
    outb(RLD | 0x10,baseAddr + 0x01);   // COUNTER TO LATCH

    outb(RLD | 0x01,baseAddr + 0x01);   // RESET BP

    outb(RESET_DETECT & 0x0000ff,baseAddr);
    outb((RESET_DETECT >> 8) & 0x0000f,baseAddr);
    outb((RESET_DETECT >> 16) & 0x0000ff,baseAddr);
}

void cioquad4_input(int port, double * y)
{
    int  enc_flags, cntrout;
    int  tmpout;
    int  tmpres = quad4[port-1].res;
    int  baseAddr = quad4[port-1].baseAddr;
    int  quad_mode = quad4[port-1].mode;
    int  firstindex = quad4[port-1].first_index;
    int  rotation = quad4[port-1].rot;
    int  counter  = quad4[port-1].count;
    double angle;
 
    enc_flags = inb(baseAddr + 0x01);  // READ FLAGS

    outb(RLD | 0x10,baseAddr + 0x01);  // RESET FLAGS
    outb(RLD | 0x01,baseAddr + 0x01);  // RESET BP

    cntrout = inb(baseAddr) & 0x00ff;     // READ COUNTER
    cntrout = cntrout | ((inb(baseAddr) & 0x00ff) * 0x100);
    cntrout = cntrout | ((inb(baseAddr) & 0x00ff) * 0x10000);

    tmpout  = cntrout;

    if ((enc_flags & 0x03) != 0x00){
	firstindex=0;
	quad4[port-1].first_index=firstindex;
	outb(IOR | 0x00,baseAddr + 0x01); // DISABLE COUNTER
    }
  
    if (!firstindex) {
//      if (enc_flags & 0x04){
	if(abs(tmpout-INDEX_DETECT)>(2*tmpres*quad_mode)){
	    firstindex=1;
	    quad4[port-1].first_index=firstindex;
         
	    if(counter == 2) outb(IOR | 0x01,baseAddr + 0x01); // ENABLE AB
	    // LOAD CNTR input
	    else             outb(IOR | 0x03,baseAddr + 0x01); // ENABLE AB
	    // LOAD OL input
	}
	else { 
	    y[0]  = 0.0;
	}
    }
    else{
	tmpout-=RESET_DETECT;
  
	angle = (double) (tmpout)/((double)tmpres*quad_mode) * 2.0 * PI;

	if (rotation == 2) {
	    angle = -angle;
	}  
	y[0]  = angle;
    }
}

void cioquad4_update()
{
}

void cioquad4_end()
{
}
