/*
COPYRIGHT (C) 2003  Roberto Bucher (roberto.bucher@die.supsi.ch)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include "asm/io.h"
#include "linux/delay.h"
#include "linux/module.h"
#include "linux/pci.h"
#include "linux/pci_ids.h"

#define SEL8800    0x0100
#define SEL7376    0x0200
#define SDI        0x8000

#define CALEN      0x4000

#define DAC0_BIP10V_COEFF   0xE0
#define DAC1_BIP10V_COEFF   0xF0

#define DAC0_FINE_GAIN     0
#define DAC0_COARSE_GAIN   1
#define DAC0_FINE_OFFSET   6
#define DAC0_COARSE_OFFSET 2
#define DAC1_FINE_GAIN     4
#define DAC1_COARSE_GAIN   5
#define DAC1_FINE_OFFSET   7
#define DAC1_COARSE_OFFSET 3

#define NVRAM_CTRL_REG    base_pci+0x3F
#define NVRAM_DATA_REG    base_pci+0x3E
#define CALIBRATE_REG     base_status+0x06

static unsigned int base_dac;

static unsigned char get_pci_dac1602_nVRam(unsigned int base_pci, unsigned int addr)
{

    unsigned char value;

    while (inb(NVRAM_CTRL_REG) & 0x80);   // D7=0 (not busy)
  
    outb(0x80, NVRAM_CTRL_REG);           // CMD to load low byte of address 
    outb(addr, NVRAM_DATA_REG);           // load low byte 

    outb(0xa0, NVRAM_CTRL_REG);           // CMD to load high byte of address 
    outb(0x00, NVRAM_DATA_REG);           // load high byte 

    outb(0xe0, NVRAM_CTRL_REG);           // CMD to read NVRAM data 
  
    while (inb(NVRAM_CTRL_REG) & 0x80);   // D7=0 (not busy) 
    value = inb(NVRAM_DATA_REG);
 
    return (value);
}

static void write_pci_dac1602_8800( unsigned int base_status, unsigned char addr, unsigned char value )
{
    int i;

    /* write 3 bits MSB first of address, keep SEL8800 low */
    for ( i = 0; i < 3; i++) {
	if ( addr & 0x4 ) {
	    outw_p(CALEN | SDI, CALIBRATE_REG );
	} else {
	    outw_p(CALEN | 0x0, CALIBRATE_REG );
	}
	udelay(10);
	addr <<= 1;
    }

    /* write 8 bits MSB first of data, keep SEL8800 low */
    for ( i = 0; i < 8; i++) {
	if ( value & 0x80 ) {
	    outw_p(CALEN | SDI, CALIBRATE_REG );
	} else {
	    outw_p(CALEN | 0x0, CALIBRATE_REG );
	}
	udelay(10);
	value <<= 1;
    }

    /* SEL8800 to set load clock */
    outw_p( SEL8800, CALIBRATE_REG );
    udelay(10);

    /* SEL8800 to reset load clock and update output */
    outw_p( 0x0, CALIBRATE_REG );
}

static void cal_da(unsigned int base_pci, unsigned int base_status)
{
    unsigned char addr,value;

    addr=DAC0_BIP10V_COEFF;

    value = get_pci_dac1602_nVRam( base_pci, addr++ );  
    write_pci_dac1602_8800 ( base_status, DAC0_COARSE_GAIN, value );
    value = get_pci_dac1602_nVRam( base_pci, addr++ );
    write_pci_dac1602_8800 ( base_status, DAC0_COARSE_OFFSET, value );
    value = get_pci_dac1602_nVRam( base_pci, addr++ );
    write_pci_dac1602_8800 ( base_status, DAC0_FINE_GAIN, value );

    addr=DAC1_BIP10V_COEFF;

    value = get_pci_dac1602_nVRam( base_pci, addr++ );
    write_pci_dac1602_8800 ( base_status, DAC1_COARSE_GAIN, value );
    value = get_pci_dac1602_nVRam( base_pci, addr++ );
    write_pci_dac1602_8800 ( base_status, DAC1_COARSE_OFFSET, value );
    value = get_pci_dac1602_nVRam( base_pci, addr++ );
    write_pci_dac1602_8800 ( base_status, DAC1_FINE_GAIN, value );
  
}

static int get_pci_addr(unsigned int vendor, unsigned int device, unsigned int addr_mat[])
{
    struct pci_dev *pdev=NULL;
    unsigned int base_pci, base_status, base_adc, base_dio, base_dac;

    pdev=pci_find_device(vendor,device,pdev);
    if(pdev) {

	pci_read_config_dword(pdev,PCI_BASE_ADDRESS_0, &base_pci);
	pci_read_config_dword(pdev,PCI_BASE_ADDRESS_1, &base_status);
	pci_read_config_dword(pdev,PCI_BASE_ADDRESS_2, &base_adc);
	pci_read_config_dword(pdev,PCI_BASE_ADDRESS_3, &base_dio);
	pci_read_config_dword(pdev,PCI_BASE_ADDRESS_4, &base_dac);

	base_pci    &= PCI_BASE_ADDRESS_IO_MASK;
	base_status &= PCI_BASE_ADDRESS_IO_MASK;
	base_adc    &= PCI_BASE_ADDRESS_IO_MASK;
	base_dio    &= PCI_BASE_ADDRESS_IO_MASK;
	base_dac    &= PCI_BASE_ADDRESS_IO_MASK;

	addr_mat[0]=base_pci;
	addr_mat[1]=base_status;
	addr_mat[2]=base_adc;
	addr_mat[3]=base_dio;
	addr_mat[4]=base_dac;
    }
    return 0;
}

void dapci1200_init()
{
    unsigned int base_status, base_pci;
    unsigned int addr_mat[5];

    get_pci_addr(PCI_VENDOR_ID_CBOARDS,0x0F,addr_mat);

    base_pci    = addr_mat[0];
    base_status = addr_mat[1];
    base_dac    = addr_mat[4];

    outw(0x0502,base_status+0x08);

    cal_da(base_pci,base_status);
}

void dapci1200_output(int channels, double * u)
{
    int i;
    int res;

    for(i=0; i<channels; i++){
	if(u[i]>10.0) u[i]=10.0;
	if(u[i]<-10.0) u[i]=-10.0;
	res=(int) ((u[i]+10.0)/20.0*4095.0);
	outw(res,base_dac+2*i);
    }
}

void dapci1200_end(int channels)
{
    int i;

    for(i=0; i<channels; i++){
	outw(0x7ff,base_dac+2*i);
    }
}
