/*
COPYRIGHT (C) 2003  Roberto Bucher (roberto.bucher@die.supsi.ch)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/string.h>
#include <asm/param.h>

#include "net_rpc.h"

#include "devices.h"

#define TASK_PRIORITY                 1
#define STACK_SIZE                    25000
#define rt_MainTaskPriority           100001
#define rt_HostInterfaceTaskPriority  100000
#define rt_BaseTaskPriority           0

#define MAX_ADR_SRCH                  500
#define POLL_PERIOD                   100000000
#define MAX_NAME_SIZE                 50
#define MAX_SCOPES                    100
#define MAX_LOGS                      100

typedef struct rtTargetParamInfo {
        char modelName[MAX_NAME_SIZE];
        char blockName[MAX_NAME_SIZE];
        char paramName[MAX_NAME_SIZE];
        unsigned int dataType;
        unsigned int dataClass;
        double dataValue;
} rtTargetParamInfo;

static RT_TASK *rt_MainTask, *rt_HostInterfaceTask, *rt_BaseRateTask;

extern void * rt_get_adr(unsigned long name);

char *TargetLogMbxID = "RTL";

static char * HostInterfaceTaskName  = "IFTASK";
MODULE_PARM(HostInterfaceTaskName, "s");
char *TargetMbxID    = "RTS";
MODULE_PARM(TargetMbxID, "s");
static int p = 100000000/HZ;
MODULE_PARM(p, "i");
static int InternTimer = 1;
MODULE_PARM(InternTimer, "i");
static int ClockTick = 0;
MODULE_PARM(ClockTick, "i");
static RTIME rt_BasePeriod = 100000000/HZ;
MODULE_PARM(rt_BasePeriod, "i");
static int isRunning = 1;
MODULE_PARM(isRunning, "i");

static volatile int endex;

static RTIME t_samp, tick_period;
static int NUPAR1       = 0;

static double TIME;
static struct { char name[MAX_NAME_SIZE]; int ntraces; } rtaiScope[MAX_SCOPES];
static struct { char name[MAX_NAME_SIZE]; int nrow, ncol; } rtaiLogData[MAX_LOGS];
static double *UPAR1;

#define SS_DOUBLE  0
#define rt_SCALAR  0

static char *ComNode = "127.0.0.1";
static unsigned long comnode;

static inline void strncpyz(char *dest, const char *src, size_t n)
{
    int i;

    for(i=0;i<n;i++) dest[i]=src[i];
    dest[n-1] = '\0';
}

void *get_a_name(const char *root, char *name)
{
        unsigned long i;
        for (i = 0; i < MAX_ADR_SRCH; i++) {
                sprintf(name, "%s%ld", root, i);
                if (!RT_get_adr(0,0,name)) {
                        return name;
                }
        }
        return 0;
}

/* static unsigned long get_an_id(const char *root) */
/* { */
/*         char name[7]; */
/*         unsigned long i, id; */
/*         for (i = 0; i < MAX_ADR_SRCH; i++) { */
/*                 sprintf(name, "%s%d", root, i); */
/*                 if (!rt_get_adr(id = nam2num(name))) { */
/*                         return id; */
/*                 } */
/*         } */
/*         return 0; */
/* } */

int rtRegisterScope(const char *name, int ntraces)
{
        int i;
        for (i = 0; i < MAX_SCOPES; i++) {
                if (!rtaiScope[i].ntraces) {
                        rtaiScope[i].ntraces = ntraces;
                        strncpy(rtaiScope[i].name, name, MAX_NAME_SIZE);
                        return 0;
                }
        }
        return 1;
}

static int rtRegisterLogData(const char *name, int nrow, int ncol)
{
    int i;
    for (i = 0; i < MAX_SCOPES; i++) {
        if (!rtaiLogData[i].nrow) {
            rtaiLogData[i].nrow = nrow;
            rtaiLogData[i].ncol = ncol;
            strncpyz(rtaiLogData[i].name, name, MAX_NAME_SIZE);
            return 0;
        }
    }
    return -1;
}

unsigned long udn2nl(double dudn)
{
        char ip[16];
        unsigned long r0, r1, r2, r3;

        if (!(r0 = dudn/1000000000.0)) {;
                return 0;
        }
        r1 = ((unsigned long)(dudn/1000000.0))%1000;
        r2 = (r3 = (unsigned long)(dudn/1000.0))%1000;
        r3 = dudn - (double)r3*1000.0;
        sprintf(ip, "%lu.%lu.%lu.%lu", r0, r1, r2, r3);
        return ddn2nl(ip);
}

static void (*WaitTimingEvent)(unsigned long);
static void (*SendTimingEvent)(unsigned long);
static unsigned long TimingEventArg;

#define XNAME(x,y) x##y
#define NAME(x,y)  XNAME(x,y)

#define XSTR(x)    #x
#define S_TR(x)     XSTR(x)
#define MODELNAME S_TR(NAME(MODEL,_standalone.c))

#include MODELNAME

static int rtModifyRParam(int i, double *param)
{
        if (i >= 0 && i < NRPAR1) {
                RPAR1[i] = *param;
                return 0;
        }
        return -1;
}

static inline int rtModifyIParam(int i, int param)
{
    if (i >= 0 && i < NIPAR1) {
        IPAR1[i] = param;
        return 0;
    }
    return -1;
}

static int rtModifyUParam(int i, double *param)
{
        if (i >= 0 && i < NRPAR1) {
                UPAR1[i] = *param;
                return 0;
        }
        return -1;
}

static int rtFillRParam(rtTargetParamInfo *rtpi, int i)
{
        if (i >= 0 && i < NRPAR1) {
                sprintf(rtpi->paramName, "RPAR1:%d", i);
                strncpyz(rtpi->modelName, MODELNAME, MAX_NAME_SIZE);
                strncpyz(rtpi->blockName, MODELNAME"/REAL PARAMS", MAX_NAME_SIZE
);
                rtpi->dataType  = SS_DOUBLE;
                rtpi->dataClass = rt_SCALAR;
                rtpi->dataValue = RPAR1[i];
                return 0;
        }
        return 1;
}

static int rtFillUParam(rtTargetParamInfo *rtpi, int i)
{
        if (i >= 0 && i < NUPAR1) {
                sprintf(rtpi->paramName, "UPAR1:%d", i);
                strncpyz(rtpi->modelName, MODELNAME, MAX_NAME_SIZE);
                strncpyz(rtpi->blockName, MODELNAME"/USER PARAMS", MAX_NAME_SIZE
);
                rtpi->dataType  = SS_DOUBLE;
                rtpi->dataClass = rt_SCALAR;
                rtpi->dataValue = UPAR1[i];
                return 0;
        }
        return 1;
}

static void rt_BaseRate(int arg)
{
    RTIME now, t0;

    NAME(MODEL,_init)(z, &TIME, RPAR1, &NRPAR1, IPAR1, &NIPAR1);

    rt_send(rt_MainTask,0);
    rt_task_suspend(rt_BaseRateTask);
    now=rt_get_time();
    t0=rt_get_cpu_time_ns();
    rt_task_make_periodic(rt_BaseRateTask,
                          now+nano2count(rt_BasePeriod),
                          nano2count(rt_BasePeriod));

    while(1){
        WaitTimingEvent((unsigned long) rt_BaseRateTask);
	TIME = (rt_get_cpu_time_ns() - t0)*1.0E-9;
	set_nevprt(1);
	NAME(MODEL,main1)(z, &TIME, RPAR1, &NRPAR1, IPAR1, &NIPAR1);
	NAME(MODEL,main2)(z, &TIME, RPAR1, &NRPAR1, IPAR1, &NIPAR1);
    }
}

static inline void modify_any_param(int index, double param)
{
    if (index < NRPAR1) {
        rtModifyRParam(index, &param);
        NAME(MODEL,_const_update)(z, &TIME, RPAR1, &NRPAR1, IPAR1, &NIPAR1);
    } else if (index < (NRPAR1 + NIPAR1)) {
        rtModifyIParam(index -= NRPAR1, (int)param);
        NAME(MODEL,_upar_update)(index);
    } else {
        rtModifyUParam(index -= (NRPAR1 + NIPAR1), &param);
        NAME(MODEL,_upar_update)(index);
    }
}

static void rt_HostInterface(int arg)
{
    RT_TASK *task;
    int iRequest;
    char Request;
    int len, Reply;

    while (1) {
	task = rt_receive(0, &iRequest);
	Request=(char) iRequest;
	switch (Request) {
	    case 'c': {
		int i, Idx;
                rtTargetParamInfo rtParam;
                float samplingTime;

                strncpyz(rtParam.modelName, MODELNAME, MAX_NAME_SIZE);
                rtParam.dataType  = SS_DOUBLE;
                rtParam.dataClass = rt_SCALAR;
                rt_return(task, (isRunning << 16) | ((NRPAR1 + NIPAR1 + NUPAR1) & 0xFFFF));
                strncpyz(rtParam.blockName, MODELNAME"/REAL PARAMS", MAX_NAME_SIZE);
                for (i = 0; i < NRPAR1; i++) {
                    rt_receivex(task, &Request, 1, &len);
                    sprintf(rtParam.paramName, "RPAR1[%d]", i);
                    rtParam.dataValue = RPAR1[i];
                    rt_returnx(task, &rtParam, sizeof(rtParam));
                }
                strncpyz(rtParam.blockName, MODELNAME"/INTEGER PARAMS", MAX_NAME_SIZE);
                for (i = 0; i < NIPAR1; i++) {
                    rt_receivex(task, &Request, 1, &len);
                    sprintf(rtParam.paramName, "IPAR1[%d]", i);
                    rtParam.dataValue = IPAR1[i];
                    rt_returnx(task, &rtParam, sizeof(rtParam));
                }
                strncpyz(rtParam.blockName, MODELNAME"/USER PARAMS", MAX_NAME_SIZE);
                for (i = 0; i < NUPAR1; i++) {
                    rt_receivex(task, &Request, 1, &len);
                    sprintf(rtParam.paramName, "UPAR1[%d]", i);
                    rtParam.dataValue = UPAR1[i];
                    rt_returnx(task, &rtParam, sizeof(rtParam));
                }
                while (1) {
                    rt_receivex(task, &Idx, sizeof(int), &len);
                    if (Idx < 0) {
                        rt_returnx(task, &Idx, sizeof(int));
                        break;
                    } else {
                        rt_returnx(task, &rtaiScope[Idx].ntraces, sizeof(int));
                        rt_receivex(task, &Idx, sizeof(int), &len);
                        rt_returnx(task, rtaiScope[Idx].name, MAX_NAME_SIZE);
                        rt_receivex(task, &Idx, sizeof(int), &len);
                        samplingTime = rt_BasePeriod*1.0E-9;
                        rt_returnx(task, &samplingTime, sizeof(float));
                    }
                }
                while (1) {
                    rt_receivex(task, &Idx, sizeof(int), &len);
                    if (Idx < 0) {
                        rt_returnx(task, &Idx, sizeof(int));
                        break;
                    } else {
                        rt_returnx(task, &rtaiLogData[Idx].nrow, sizeof(int));
                        rt_receivex(task, &Idx, sizeof(int), &len);
                        rt_returnx(task, &rtaiLogData[Idx].ncol, sizeof(int));
                        rt_receivex(task, &Idx, sizeof(int), &len);
                        rt_returnx(task, rtaiLogData[Idx].name, MAX_NAME_SIZE);
                        rt_receivex(task, &Idx, sizeof(int), &len);
                        samplingTime = rt_BasePeriod*1.0E-9;
                        rt_returnx(task, &samplingTime, sizeof(float));
                    }
                }
		break;
	    }

	    case 's': {
		Reply = 1;
		rt_task_resume(rt_BaseRateTask);
		isRunning=1;
		rt_return(task, Reply);
		break;
	    }

	    case 't': {
		endex=1;
		Reply = 0;
		rt_return(task, Reply);
		break;
	    }

	    case 'p': {
                int index;
                double param;

                rt_return(task, isRunning);
                rt_receivex(task, &index, sizeof(int), &len);
                Reply = 0;
                rt_returnx(task, &Reply, sizeof(int));
                rt_receivex(task, &param, sizeof(double), &len);
                modify_any_param(index, param);
                Reply = 1;
                rt_returnx(task, &Reply, sizeof(int));
		break;
	    }

                       case 'g': {
                                int i;
                                double value;
                                rt_return(task, isRunning);
                                for (i = 0; i < NRPAR1; i++) {
                                        rt_receivex(task, &Request, 1, &len);
                                        rt_returnx(task, &RPAR1[i], sizeof(double));
                                }
                                for (i = 0; i < NIPAR1; i++) {
                                        rt_receivex(task, &Request, 1, &len);
                                        value = IPAR1[i];
                                        rt_returnx(task, &value, sizeof(double))
;
                                }
                                for (i = 0; i < NUPAR1; i++) {
                                        rt_receivex(task, &Request, 1, &len);
                                        rt_returnx(task, &UPAR1[i], sizeof(double));
                                }
                                break;
                        }

	    case 'd': {
                int ParamCnt;
                rt_return(task, isRunning);
                rt_receivex(task, &ParamCnt, sizeof(int), &len);
                Reply = 0;
                rt_returnx(task, &Reply, sizeof(int));
                {
                    struct {
                        int index;
                        double value;
                    } Params[ParamCnt];
                    int i;
                    rt_receivex(task, &Params, sizeof(Params), &len);
                    for (i = 0; i < ParamCnt; i++) {
                        modify_any_param(Params[i].index, Params[i].value);
                    }
                }
                Reply = 1;
                rt_returnx(task, &Reply, sizeof(int));
		break;
	    }

	    case 'm': {
		float time = TIME;
		Reply = isRunning;
		rt_return(task, Reply);
		rt_receivex(task, &Reply, sizeof(int), &len);
		rt_returnx(task,&time,sizeof(float));
		break;
	    }

            case 'b': {
                rt_return(task, (unsigned int)rt_BaseRateTask);
                break;
            }

	    default: {
                rt_return(task, 0xFFFFFFFF);
		break;
	    }
	}
    }
}

static void rt_Main(int arg)
{
    SEM * hard_timers_cnt;

    t_samp=rt_BasePeriod;
    if(!(hard_timers_cnt=rt_get_adr(nam2num("HTMRCN")))){
        tick_period=start_rt_timer(nano2count(t_samp));
        hard_timers_cnt=RT_typed_named_sem_init(0,0,"HTMRCN",0,CNT_SEM);
    }
    else{
        RT_sem_signal(0,0,hard_timers_cnt);
    }
    {
	int msg;
	rt_receive(0, &msg);
    }
    while(!endex){
	rt_sleep(nano2count(POLL_PERIOD));
    }
    if(!rt_sem_wait_if(hard_timers_cnt)){
        stop_rt_timer();
        RT_named_sem_delete(0,0,hard_timers_cnt);
    }
}

int start_tasks()
{
    char name[7]="MNT";

    rt_BasePeriod = p;

    strncat(name,TargetMbxID,3);
    if (!(rt_MainTask = RT_named_task_init(0, 0, name, rt_Main, 0, STACK_SIZE,
					   rt_MainTaskPriority, 1, 0))) {
	rt_printk("Cannot init rt_MainTask.\n");
	return 1;
    }
    endex=0;
    rt_task_resume(rt_MainTask);
    rt_printk("TARGET STARTS.\n");

    if (!(rt_HostInterfaceTask = RT_named_task_init(0,0,
						    HostInterfaceTaskName,
						    rt_HostInterface, 0,
                                                    STACK_SIZE,
						    rt_HostInterfaceTaskPriority,
						    1, 0))) {
	rt_printk("Cannot init rt_HostInterfaceTask.\n");
	return 1;
    }
    rt_task_resume(rt_HostInterfaceTask);

    strcpy(name,"BRT");
    strncat(name,TargetMbxID,3);
    if (!(rt_BaseRateTask = RT_named_task_init(0,0, name,
					       rt_BaseRate, 0,
                                               STACK_SIZE,
					       rt_BaseTaskPriority, 0, 0))) {
	rt_printk("Cannot init rt_BaseRateTask.\n");
	return 1;
    }
    rt_task_use_fpu(rt_BaseRateTask,1);
    rt_task_resume(rt_BaseRateTask);

    WaitTimingEvent = (void *)rt_task_wait_period;

    rt_printk("rt_process loaded with period %ld\n",rt_BasePeriod);
    rt_printk("Target is running.\n");
    if(isRunning) rt_task_resume(rt_BaseRateTask); 

    return(0);
}


int init_module(void)
{
    return (start_tasks());
}

void cleanup_module(void)
{
    NAME(MODEL,_end)(z, &TIME, RPAR1, &NRPAR1, IPAR1, &NIPAR1);
    RT_named_task_delete(0, 0, rt_HostInterfaceTask);
    RT_named_task_delete(0, 0, rt_BaseRateTask);
    RT_named_task_delete(0,0,rt_MainTask);
    rt_printk("TARGET ENDS\n");
    rt_printk("rt_process unloaded\n");
}

MODULE_LICENSE("GPL");
