/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include "scopecanvas.h"

ScopeCanvas::ScopeCanvas(int x, int y, int w, int h, const char *title):Fl_Gl_Window(x,y,w,h,title)
{
	int i;

	BackGroundColor[0] = 0.0;
	BackGroundColor[1] = 0.0;
	BackGroundColor[2] = 0.0;
	GridColor[0] = 0.;
	GridColor[1] = 100./255.;
	GridColor[2] = 0.;
	GridStyle = 1;
	dxGrid = w/(float)((int)(w/60.));
	dyGrid = h/2;

	TRange = 1.0;
	for (i = 0; i < MAX_CHANNELS; i++) {
		UnitsDiv[i] = 2.5;
		YRangeMin[i] = -UnitsDiv[i]*4.;
		YRangeMax[i] = UnitsDiv[i]*4.;
		TraceOffset[i] = h/2;
		TraceYmin[i] = 0.;
		TraceYmax[i] = 0.;
		TraceIsVisible[i] = 1;
		TraceColor[0][i] = 1.0;
		TraceColor[1][i] = 1.0;
		TraceColor[2][i] = 1.0;
	}

	ctrFirstCall = 1;
	cysFirstCall = 1;
	ScopePauseFlag = 0;
	OneShotFlag = 0;
	StopRedraw = 0;
}

void ScopeCanvas::setSamplingFrequency(float freq)
{
	SampFreq = freq;
}

float ScopeCanvas::SamplingFrequency()
{
	return SampFreq;
}

void ScopeCanvas::setTimeRange(float range)
{
	TRange = range;
}

float ScopeCanvas::TimeRange()
{
	return TRange;
}

void ScopeCanvas::setdx()
{
	dx = w()/(SampFreq*TRange);
}

float ScopeCanvas::getdx()
{
	return dx;
}

void ScopeCanvas::setTraceLength()
{
	TraceLen = (int)(w()/dx);
}

int ScopeCanvas::TraceLength()
{
	return TraceLen;
}

void ScopeCanvas::setTracePointer(int pos)
{
	TracePtr = pos;
}

int ScopeCanvas::TracePointer()
{
	return TracePtr;
}

int ScopeCanvas::incrementTracePointer()
{
	return TracePtr++;
}

float ScopeCanvas::minTraceValue(int n)
{
	return TraceYmin[n];
}

float ScopeCanvas::maxTraceValue(int n)
{
	return TraceYmax[n];
}

void ScopeCanvas::setYmin(float val, int n)
{
	YRangeMin[n] = val;
}

void ScopeCanvas::setYmax(float val, int n)
{
	YRangeMax[n] = val;
}

float ScopeCanvas::Ymin(int n)
{
	return YRangeMin[n];
}

float ScopeCanvas::Ymax(int n)
{
	return YRangeMax[n];
}

void ScopeCanvas::fillTrace(int pos, int n, float val)
{
	Trace[pos][n] = val;
}

void ScopeCanvas::initgl()
{
	glViewport(0, 0, w(), h());
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, w(), 0, h(), -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(BackGroundColor[0], BackGroundColor[1], BackGroundColor[2], 0.0);
	glClear(GL_COLOR_BUFFER_BIT);	
}	

void ScopeCanvas::drawgrid()
{
	glLineWidth(0.1);
	if (GridStyle) {
		glLineStipple(1, 0xAAAA);
		glEnable(GL_LINE_STIPPLE);
	}
	glColor3f(GridColor[0], GridColor[1], GridColor[2]);
	glBegin(GL_LINES);
	glVertex2f(0., 0.);
	glVertex2f(w(), 0.);
	float y = 0.;
	for (int i = 1; i < (int)(h()); i++) {
		y += dyGrid;
		glVertex2f(0., y);
		glVertex2f(w(), y);
	}
	float x = 0.;
	for (int i = 1; i < (int)(w()); i++) {
		x += dxGrid;
		glVertex2f(x, 0.);
		glVertex2f(x, h());
	}
	glEnd();
	if (GridStyle) {
		glDisable(GL_LINE_STIPPLE);
	}
}

void ScopeCanvas::draw()
{
	if(!valid()) {
		initgl();
		dxGrid = w()/(float)((int)(w()/40.));
		dyGrid = (h()/2.)/(float)((int)((h()/2.)/40.));
	}
	dx = w()/(SampFreq*TRange);
	TraceLen = (int)(w()/dx);
	glClearColor(BackGroundColor[0], BackGroundColor[1], BackGroundColor[2], 0.0);
	glClear(GL_COLOR_BUFFER_BIT);
	drawgrid();
	glLineWidth(0.1);
	for (int nn = 0; nn < nTraces; nn++) {
		glColor3f(TraceColor[0][nn], TraceColor[1][nn], TraceColor[2][nn]);
		if (TraceIsVisible[nn]) {
			glBegin(GL_LINE_STRIP);
			rtPoint p;
			p.x = p.y = 0;
			for (int n = 0; n < TracePtr; n++) {
				p.y = Trace[n][nn]*(((float)(h()))/(float)(YRangeMax[nn]-YRangeMin[nn])) + TraceOffset[nn];
				glVertex2f(p.x, p.y);
				p.x += dx;
				if (Trace[n][nn] > TraceYmax[nn]) TraceYmax[nn] = Trace[n][nn];
				if (Trace[n][nn] < TraceYmin[nn]) TraceYmin[nn] = Trace[n][nn];
			}
			glEnd();
		}
	}
}
