/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#ifndef _SCOPECANVAS_H_
#define _SCOPECANVAS_H_

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Preferences.H>
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Float_Input.H>
#include <FL/filename.H>
/*
#include <FL/Fl_Input.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Widget.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Light_Button.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Scroll.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Secret_Input.H>
#include <FL/Fl_Dial.H>
#include <FL/Fl_Color_Chooser.H>
#include <FL/forms.H>
*/
#include <FL/gl.h>
#include "rtailab.h"

class ScopeCanvas : public Fl_Gl_Window
{
	public:
		ScopeCanvas(int x, int y, int w, int h, const char *title="rtScope");

		void setSamplingFrequency(float freq);
		float SamplingFrequency();
		void setTimeRange(float range);
		float TimeRange();
		void setdx();
		float getdx();
		void setTraceLength();
		int TraceLength();
		void setTracePointer(int pos);
		int TracePointer();
		int incrementTracePointer();

		float minTraceValue(int n);
		float maxTraceValue(int n);
		void setYmin(float val, int n);
		float Ymin(int n);
		void setYmax(float val, int n);
		float Ymax(int n);
		void fillTrace(int pos, int n, float val);

		char *name;
		int nTraces;
		float UnitsDiv[MAX_CHANNELS];
		float TraceOffset[MAX_CHANNELS];
		int TraceIsVisible[MAX_CHANNELS];

		Fl_Box *SecDivLabel;
		Fl_Box *SecDivLabel2;
		Fl_Box *UnitsDivLabel[MAX_CHANNELS];
		Fl_Box *UnitsDivLabel2[MAX_CHANNELS];

		char *y1name, *y2name;
		Fl_Window *crtWin;
		Fl_Float_Input *TimeRangeInput;
		Fl_Preferences *trange;
		Fl_Box *x0label, *xflabel;
		int ctrFirstCall;
		Fl_Window *cysWin;
		Fl_Float_Input *YminInput, *YmaxInput;
		Fl_Preferences *yminv, *ymaxv;
		Fl_Box *yminlabel, *ymaxlabel;
		int cysFirstCall;

		float BackGroundColor[3];
		float GridColor[3];
		float TraceColor[3][MAX_CHANNELS];
		int GridStyle;
		int ScopePauseFlag;
		int OneShotFlag;
		int StopRedraw;
	private:
		float SampFreq;
		float TRange;
		float YRangeMin[MAX_CHANNELS];
		float YRangeMax[MAX_CHANNELS];
		float dx;
		int TraceLen, TracePtr;
		float TraceYmin[MAX_CHANNELS];
		float TraceYmax[MAX_CHANNELS];
		float Trace[MAX_TRACE_LENGTH][MAX_CHANNELS];
	protected:
		void initgl();
		void drawgrid();
		void draw();
		float dxGrid;
		float dyGrid;
};

#endif
