/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include <FL/forms.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>
#include "scopepanel.h"
#include "controlpanel.h"
#include "rtailab.h"
#include <string.h>

extern int NumRtScopes;
extern int NumTunableParameters;
extern rtTargetParamInfo *TunableParameters;
extern rtTargetScopeInfo RtScopes[];
extern ControlPanel *CommandManager;

void ScopePanel::show()
{
        Window->show();
}

void ScopePanel::hide()
{
        Window->hide();
}

int ScopePanel::IsLoggingEnabled(int i)
{
	return LogFlag[i];
}

inline void ScopePanel::cb_plot_i(Fl_Check_Button *o, void *v)
{
	int n = (int)v;
	o->value() ? RtScopes[n].visible = 1 : RtScopes[n].visible = 0;
}

void ScopePanel::cb_plot(Fl_Check_Button *o, void *v)
{
	((ScopePanel *)(o->parent()->parent()->parent()->parent()->user_data()))->cb_plot_i(o,v);
}

inline void ScopePanel::cb_log_i(Fl_Check_Button *o, void *v)
{
	int n = (int)v;
	if (o->value()) {
		if ((LogFile[n] = fopen(LogFileName[n], "a+")) == NULL) {
			fl_alert("Error in opening file %s", LogFileName[n]);
			o->value(0);
			return;
		}
		LogFlag[n] = true;
	} else {
		if (LogFile[n]) {
			fclose(LogFile[n]);
		}
		LogFlag[n] = false;
	}
}

void ScopePanel::cb_log(Fl_Check_Button *o, void *v)
{
	((ScopePanel *)(o->parent()->parent()->parent()->parent()->user_data()))->cb_log_i(o,v);
}

inline void ScopePanel::cb_close_i(Fl_Button *o, void *v)
{
	Window->hide();
	CommandManager->activate_RtPlots();
}

void ScopePanel::cb_close(Fl_Button *o, void *v)
{
	((ScopePanel *)(o->parent()->parent()->parent()->user_data()))->cb_close_i(o,v);
}

ScopePanel::ScopePanel()
{
	Fl_Window *w;
	{ Fl_Window *o = Window = new Fl_Window(250, 400);
	  w = o;
	  o->label("Scope Panel");
	  o->user_data((void *)(this));
	  { Fl_Group *o = new Fl_Group(0, 0, 250, 400);
	    { Fl_Group *o = ScopeInfo = new Fl_Group(0, 0, 250, 400);
	      { Fl_Button *o = new Fl_Button(5, 10, 90, 35, "Model name: ");
		o->labelsize(12);
		o->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
		o->box(FL_NO_BOX);
	      }
	      { Fl_Button *o = new Fl_Button(100, 10, 90, 35, strdup(TunableParameters[0].modelName));
		o->labelsize(14);
		o->labeltype(FL_EMBOSSED_LABEL);
		o->labelfont(3);
		o->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
		o->box(FL_NO_BOX);
	      }
	      { Fl_Scroll *o = ScopeScroll = new Fl_Scroll(5, 60, 245, 280);
		o->type(Fl_Scroll::VERTICAL_ALWAYS);
		o->box(FL_EMBOSSED_BOX);
		Plot = new Fl_Check_Button*[NumRtScopes];
		Log = new Fl_Check_Button*[NumRtScopes];
		LogFileName = new char*[NumRtScopes];
		LogFlag = new int[NumRtScopes];
		for (int i = 0; i < NumRtScopes; i++) {
			Fl_Box *box = new Fl_Box(FL_FRAME_BOX, 5,  60 + 35*i, 225, 35, 0);
			box->color(FL_GRAY);
			{ Fl_Check_Button *o = Plot[i] = new Fl_Check_Button(10, 65 + 35*i, 90, 25, strdup(RtScopes[i].name));
			  o->labelsize(12);
			  o->callback((Fl_Callback *)cb_plot, (void *)i);
			}
			{ Fl_Check_Button *o = Log[i] = new Fl_Check_Button(110, 65 + 35*i, 90, 25, "save to file");
                	  o->labelsize(12);
			  o->callback((Fl_Callback *)cb_log, (void *)i);
			}
                	LogFileName[i] = strdup(RtScopes[i].name);
        	}
		LogFile = new FILE*[NumRtScopes];
		ScopeScroll->end();
	      }
	      { Fl_Button *o = new Fl_Button(175, 370, 70, 25, "Close");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_close);
	      }
	      o->end();
	    }
	    o->end();
          }
	}
};
