/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include "scopewindow.h"

void ScopeWindow::getTargetTime(float time)
{
	char buf[20];
	sprintf(buf, "%1.1f", time);
	TargetTime->label(strdup(buf));
}

static void ChangeTimeRange(Fl_Valuator *w, void *v)
{
	ScopeCanvas *o = (ScopeCanvas *)v;
	Fl_Valuator *d = (Fl_Valuator *)w;
	int val = (int)(d->value());
	float tval = 0.1;
	char lbl[10];
	
	switch (val) {
		case 0:
			tval = 5.0;
			sprintf(lbl, "%gs", tval);
			break;
		case 1:
			tval = 2.5;
			sprintf(lbl, "%gs", tval);
			break;
		case 2:
			tval = 1.0;
			sprintf(lbl, "%gs", tval);
			break;
		case 3:
			tval = 0.5;
			sprintf(lbl, "%gs", tval);
			break;
		case 4:
			tval = 0.25;
			sprintf(lbl, "%gs", tval);
			break;
		case 5:
			tval = 0.1;
			sprintf(lbl, "%gs", tval);
			break;
		case 6:
			tval = 0.05;
			sprintf(lbl, "%gs", tval);
			break;
		case 7:
			tval = 0.025;
			sprintf(lbl, "%gs", tval);
			break;
		case 8:
			tval = 0.01;
			sprintf(lbl, "%gs", tval);
			break;
		case 9:
			tval = 0.005;
			sprintf(lbl, "%gs", tval);
			break;
		case 10:
			tval = 0.0025;
			sprintf(lbl, "%gs", tval);
			break;
		default:
			break;
	}
			
	o->setTimeRange(tval*10.0);
//	o->SecDivLabel->label(strdup(lbl));
	o->SecDivLabel2->label(strdup(lbl));
}

static void PauseCb(Fl_Widget *w, void *v)
{
	Fl_Check_Button *m = (Fl_Check_Button *)w;
	ScopeCanvas *o = (ScopeCanvas *)v;

	if (m->value()) {
		o->ScopePauseFlag = 1;
	} else {
		o->ScopePauseFlag = 0;
	}
}

static void OneShotCb(Fl_Widget *w, void *v)
{
	Fl_Check_Button *m = (Fl_Check_Button *)w;
	ScopeCanvas *o = (ScopeCanvas *)v;

	if (m->value()) {
		o->OneShotFlag = 1;
	} else {
		o->OneShotFlag = 0;
		o->StopRedraw = 0;
	}
}

static void ShowHideCb(Fl_Widget *w, void *v)
{
	Fl_Check_Button *m = (Fl_Check_Button *)w;
	LocalMsg *rmsg = (LocalMsg *)v;
	ScopeCanvas *o = (ScopeCanvas *)rmsg->s;
	int n = (int)rmsg->n;

	if (m->value()) {
		o->TraceIsVisible[n] = 1;
	} else {
		o->TraceIsVisible[n] = 0;
	}
}

static void AutoScaleCb(Fl_Widget *w, void *v)
{
	LocalMsg *rmsg = (LocalMsg *)v;
//	ScopeCanvas *o = (ScopeCanvas *)rmsg->s;
	int n = (int)rmsg->n;

	printf("Trace %d\n", n);
}

static void BgColorCb(Fl_Widget *w, void *v)
{
	ScopeCanvas *o = (ScopeCanvas *)v;
	Fl_Color bgc = FL_FREE_COLOR;
	uchar r,g,b;

	Fl::get_color(bgc,r,g,b);
	if (!fl_color_chooser("Bg color:",r,g,b)) return;
	Fl::set_color(bgc,r,g,b);
	o->BackGroundColor[0] = (float)r/255.;
	o->BackGroundColor[1] = (float)g/255.;
	o->BackGroundColor[2] = (float)b/255.;
}

static void GridColorCb(Fl_Widget *w, void *v)
{
	ScopeCanvas *o = (ScopeCanvas *)v;
	Fl_Color grc = FL_GREEN;
	uchar r,g,b;

	Fl::get_color(grc,r,g,b);
	if (!fl_color_chooser("Grid color:",r,g,b)) return;
	Fl::set_color(grc,r,g,b);
	o->GridColor[0] = (float)r/255.;
	o->GridColor[1] = (float)g/255.;
	o->GridColor[2] = (float)b/255.;
}


static void TraceColorCb(Fl_Widget *w, void *v)
{
	LocalMsg *rmsg = (LocalMsg *)v;
	ScopeCanvas *o = (ScopeCanvas *)rmsg->s;
	int n = (int)rmsg->n;
	uchar r,g,b;
	Fl_Color trc[MAX_CHANNELS];

	trc[n] = FL_WHITE;
	Fl::get_color(trc[n],r,g,b);
	if (!fl_color_chooser("Trace color:",r,g,b)) return;
	Fl::set_color(trc[n],r,g,b);
	o->TraceColor[0][n] = (float)r/255.;
	o->TraceColor[1][n] = (float)g/255.;
	o->TraceColor[2][n] = (float)b/255.;
}

/*
			printf("%g %g\n", o->minTraceValue(n), o->maxTraceValue(n));
			printf("%g %g\n", o->Ymin(n), o->Ymax(n));
			printf("%g %g\n", o->UnitsDiv[n], o->UnitsDiv[n]);
			o->setYmin(o->minTraceValue(n), n);
			o->setYmax(o->maxTraceValue(n), n);
*/

static void SelectYRange(Fl_Widget *w, void *v)
{
	Fl_Valuator *d = (Fl_Valuator *)w;
	LocalMsg *rmsg = (LocalMsg *)v;
	ScopeCanvas *o = (ScopeCanvas *)rmsg->s;
	int n = (int)rmsg->n;
	int val = (int)(d->value());
	float tval = 1.0;
	char lbl[10];
	
	switch (val) {
		case 0:
			tval = 50.0;
			sprintf(lbl, "%g", tval);
			break;
		case 1:
			tval = 25.0;
			sprintf(lbl, "%g", tval);
			break;
		case 2:
			tval = 10.0;
			sprintf(lbl, "%g", tval);
			break;
		case 3:
			tval = 5.0;
			sprintf(lbl, "%g", tval);
			break;
		case 4:
			tval = 2.5;
			sprintf(lbl, "%g", tval);
			break;
		case 5:
			tval = 1.0;
			sprintf(lbl, "%g", tval);
			break;
		case 6:
			tval = 0.5;
			sprintf(lbl, "%g", tval);
			break;
		case 7:
			tval = 0.25;
			sprintf(lbl, "%g", tval);
			break;
		case 8:
			tval = 0.1;
			sprintf(lbl, "%g", tval);
			break;
		case 9:
			tval = 0.05;
			sprintf(lbl, "%g", tval);
			break;
		case 10:
			tval = 0.025;
			sprintf(lbl, "%g", tval);
			break;
		default:
			break;
	}
			
	o->UnitsDiv[n] = tval;
	o->setYmin((float)(-tval*4), n);
	o->setYmax((float)(tval*4), n);
	o->UnitsDivLabel2[n]->label(strdup(lbl));
}

static void ChangeTracePos(Fl_Widget *w, void *v)
{
	Fl_Valuator *d = (Fl_Valuator *)w;
	LocalMsg *rmsg = (LocalMsg *)v;
	ScopeCanvas *o = (ScopeCanvas *)rmsg->s;
	int n = (int)rmsg->n;
	o->TraceOffset[n] = ((o->h())/2.)*d->value();
}

/*
static void AutoScaleY(Fl_Widget *, void *v)
{
	ScopeCanvas *o = (ScopeCanvas *)v;
	char buf[10];

	o->setYmin(o->minTraceValue()*1.2);
	o->setYmax(o->maxTraceValue()*1.2);
	sprintf(buf, "%1.1f", o->Ymin());
	o->yminlabel->label(strdup(buf));
	o->YminInput->value(strdup(buf));
	sprintf(buf, "%1.1f", o->Ymax());
	o->ymaxlabel->label(strdup(buf));
	o->YmaxInput->value(strdup(buf));
}
*/

ScopeWindow::~ScopeWindow()
{
	delete(Scope);
}

void ScopeWindow::show()
{
	scpWin->show();
}

void ScopeWindow::hide(void)
{
	scpWin->hide();
}

ScopeWindow::ScopeWindow(int n, char *name, int ntr, float samp_freq)
{
	Fl_Window *w;
	{ Fl_Window *o = scpWin = new Fl_Window(520 + 100*ntr, 340);
	  scpWin->label(strdup(name));
	  w = o;
	  o->color(FL_GRAY-4);

	  { Fl_Group *o = PlotArea = new Fl_Group(0, 0, 445, 360);
	    { Fl_Box *o = PlotBox = new Fl_Box(FL_EMBOSSED_FRAME, 8, 8, 404, 324, 0);
	      o->color(255);
	    }
	    { ScopeCanvas* o = Scope = new ScopeCanvas(10, 10, 400, 320, strdup(name));
	      o->name = strdup(name);
	      o->nTraces = ntr;
	      o->setSamplingFrequency(samp_freq);
	      o->setdx();
	      o->setTraceLength();
/*
	      char lab[10] = "0.1s";
	      o->SecDivLabel = new Fl_Box(FL_NO_BOX, 230, rtScopeUI_height-20, 10, 10, strdup(lab));
	      o->SecDivLabel->labelcolor(FL_BLACK);
	      o->SecDivLabel->labelsize(10);
*/
            }
	    o->end();
	  }

	  { Fl_Group *o = CommonArea = new Fl_Group(0, 0, rtScopeUI_width, rtScopeUI_height);
	    { Fl_Box *o = TargetTimeLabel = new Fl_Box(FL_NO_BOX, 445, 10, 50, 20, "Target Time [s]");
	      o->labelcolor(FL_BLACK);
	      o->labelsize(10);
	    }
	    { Fl_Box *o = TargetTime = new Fl_Box(FL_NO_BOX, 445, 30, 50, 20, 0);
	      o->labelcolor(FL_BLACK);
	      o->labelsize(10);
	    }
	    { Fl_Box *o = CommonBox = new Fl_Box(FL_FRAME_BOX, 425, 80, 90, 252, 0);
	      o->color(FL_GRAY-4);
	    }
	    { Fl_Dial *o = SecDivDial = new Fl_Dial(440, 100, 60, 60, "SEC / DIV");
	      o->type(1);
	      o->tooltip("Seconds per division");
	      o->labelsize(10);
	      o->minimum(0);
	      o->maximum(10);
	      o->step(1);
	      o->value(5);
	      o->callback((Fl_Callback *)ChangeTimeRange, (void *)Scope);
	      char lbl[10] = "0.1s";
	      Scope->SecDivLabel2 = new Fl_Box(FL_NO_BOX, 435, 180, 70, 20, strdup(lbl));
	      Scope->SecDivLabel2->labelcolor(FL_BLACK);
	      Scope->SecDivLabel2->labelsize(12);
	      Scope->SecDivLabel2->labelfont(FL_BOLD);
	    }	
	    { Fl_Check_Button *o = PauseButton = new Fl_Check_Button(435, 220, 70, 20, "Pause");
	      o->labelsize(10);
	      o->align(FL_ALIGN_CENTER);
	      o->callback((Fl_Callback *)PauseCb, (void *)Scope);
	    }
	    { Fl_Check_Button *o = OneShotButton = new Fl_Check_Button(435, 240, 70, 20, "OneShot");
	      o->labelsize(10);
	      o->align(FL_ALIGN_CENTER);
	      o->callback((Fl_Callback *)OneShotCb, (void *)Scope);
	    }
	    { Fl_Button *o = BgColorButton = new Fl_Button(435, 270, 70, 20, "Bg Color");
	      o->labelsize(10);
	      o->align(FL_ALIGN_CENTER);
	      o->callback((Fl_Callback *)BgColorCb, (void *)Scope);
	    }
	    { Fl_Button *o = GridColorButton = new Fl_Button(435, 300, 70, 20, "Grid Color");
	      o->labelsize(10);
	      o->align(FL_ALIGN_CENTER);
	      o->callback((Fl_Callback *)GridColorCb, (void *)Scope);
	    }
	    o->end();
	  }

	  for (int i = 0; i < ntr; i++) {
	  	{ Fl_Group *o = ChannelArea[i] = new Fl_Group(0, 0, rtScopeUI_width, rtScopeUI_height);
	  	  { char lbl[10];
		    sprintf(lbl, "Channel %d", i+1);
		    Fl_Box *o = ChannelBox[i] = new Fl_Box(FL_FRAME_BOX, 520 + 95*i, 30, 90, 302, strdup(lbl));
	  	    o->color(FL_GRAY-4);
		    o->align(FL_ALIGN_TOP);
		    o->labelfont(FL_BOLD);
	  	  }
	  	  { Fl_Dial *o = PosDial[i] = new Fl_Dial(540 + 95*i, 50, 50, 50, "Position");
	  	    o->type(1);
	  	    o->tooltip("Trace position on the screen");
	  	    o->labelsize(10);
	  	    o->minimum(0.0);
	  	    o->maximum(2.0);
	  	    o->value(1.0);
	  	    LocalMsg *PosMsg = (LocalMsg *)malloc(sizeof(LocalMsg));
	  	    PosMsg->s = (ScopeCanvas *)malloc(sizeof(ScopeCanvas *));
	  	    PosMsg->s = (ScopeCanvas *)Scope;
	  	    PosMsg->n = i;
	  	    o->callback((Fl_Callback *)ChangeTracePos, (void *)PosMsg);
	  	  }
	  	  { Fl_Dial *o = UnitsDivDial[i] = new Fl_Dial(535 + 95*i, 130, 60, 60, "UNITS / DIV");
	  	    o->type(1);
	  	    o->tooltip("Units per division");
	  	    o->labelsize(10);
	  	    o->minimum(0);
	  	    o->maximum(10);
	  	    o->step(1);
	  	    o->value(4);
	  	    LocalMsg *UnitsDivMsg = (LocalMsg *)malloc(sizeof(LocalMsg));
	  	    UnitsDivMsg->s = (ScopeCanvas *)malloc(sizeof(ScopeCanvas *));
	  	    UnitsDivMsg->s = (ScopeCanvas *)Scope;
	  	    UnitsDivMsg->n = i;
	  	    o->callback((Fl_Callback *)SelectYRange, (void *)UnitsDivMsg);
	            char *lbl = "2.5";
	            Scope->UnitsDivLabel2[i] = new Fl_Box(FL_NO_BOX, 530 + 95*i, 210, 70, 20, strdup(lbl));
	            Scope->UnitsDivLabel2[i]->labelcolor(FL_BLACK);
	            Scope->UnitsDivLabel2[i]->labelsize(12);
	            Scope->UnitsDivLabel2[i]->labelfont(FL_BOLD);
	  	  }
	          { Fl_Check_Button *o = ShowButton[i] = new Fl_Check_Button(530 + 95*i, 240, 70, 20, "Show/Hide");
		    o->value(1);
	            o->labelsize(10);
	            o->align(FL_ALIGN_CENTER);
	  	    LocalMsg *ShowHideMsg = (LocalMsg *)malloc(sizeof(LocalMsg));
	  	    ShowHideMsg->s = (ScopeCanvas *)malloc(sizeof(ScopeCanvas *));
	  	    ShowHideMsg->s = (ScopeCanvas *)Scope;
	  	    ShowHideMsg->n = i;
	            o->callback((Fl_Callback *)ShowHideCb, (void *)ShowHideMsg);
	          }
	          { Fl_Button *o = AutoScaleButton[i] = new Fl_Button(530 + 95*i, 270, 70, 20, "Autoscale");
	            o->labelsize(10);
	            o->align(FL_ALIGN_CENTER);
	  	    LocalMsg *AutoScaleMsg = (LocalMsg *)malloc(sizeof(LocalMsg));
	  	    AutoScaleMsg->s = (ScopeCanvas *)malloc(sizeof(ScopeCanvas *));
	  	    AutoScaleMsg->s = (ScopeCanvas *)Scope;
	  	    AutoScaleMsg->n = i;
	            o->callback((Fl_Callback *)AutoScaleCb, (void *)AutoScaleMsg);
		    o->deactivate();
	          }
	          { Fl_Button *o = TraceColorButton[i] = new Fl_Button(530 + 95*i, 300, 70, 20, "Color");
	            o->labelsize(10);
	            o->align(FL_ALIGN_CENTER);
	  	    LocalMsg *TrColorMsg = (LocalMsg *)malloc(sizeof(LocalMsg));
	  	    TrColorMsg->s = (ScopeCanvas *)malloc(sizeof(ScopeCanvas *));
	  	    TrColorMsg->s = (ScopeCanvas *)Scope;
	  	    TrColorMsg->n = i;
	            o->callback((Fl_Callback *)TraceColorCb, (void *)TrColorMsg);
	          }
	  	  o->end();
	  	}
	  }
	}
}
