/*
COPYRIGHT (C) 2003  Lorenzo Dozio (dozio@aero.polimi.it)
                    Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include "controlpanel.h"
#include "targetpanel.h"
#include <stdlib.h>

extern Fl_Preferences app;
extern ControlPanel *CommandManager;

char *TargetPanel::getIP()
{
	char *IPvalue;
	IPaddrPref->get("targetIP", IPvalue, "127.0.0.1");
	return IPvalue;
}

char *TargetPanel::getIfTaskID()
{
	char *IDvalue;
	IfTaskIDPref->get("targetIf", IDvalue, "IFTASK");
	return IDvalue;
}

char *TargetPanel::getScopeID()
{
	char *IDvalue;
	ScopeIDPref->get("targetScopeMbx", IDvalue, "RTS");
	return IDvalue;
}

char *TargetPanel::getLogID()
{
	char *IDvalue;
	LogIDPref->get("targetLogMbx", IDvalue, "RTL");
	return IDvalue;
}

void TargetPanel::show()
{
	Window->show();
}

void TargetPanel::hide()
{
	Window->hide();
}

inline void TargetPanel::cb_ok_i(Fl_Button *o, void *v)
{
	IPaddrPref->set("targetIP", IPaddress->value());
	IfTaskIDPref->set("targetIf", InterfaceTaskID->value());
	ScopeIDPref->set("targetScopeMbx", ScopeMbxID->value());
	LogIDPref->set("targetLogMbx", LogMbxID->value());
	Window->hide();
	CommandManager->activate_SelectTarget();
}

void TargetPanel::cb_ok(Fl_Button *o, void *v)
{
	((TargetPanel *)(o->parent()->parent()->parent()->user_data()))->cb_ok_i(o,v);
}

inline void TargetPanel::cb_cancel_i(Fl_Button *o, void *v)
{
	Window->hide();
	CommandManager->activate_SelectTarget();
}

void TargetPanel::cb_cancel(Fl_Button *o, void *v)
{
	((TargetPanel *)(o->parent()->parent()->parent()->user_data()))->cb_cancel_i(o,v);
}

TargetPanel::TargetPanel()
{
	IPaddrPref = new Fl_Preferences(app, "Network");
	IfTaskIDPref = new Fl_Preferences(app, "TgIfID");
	ScopeIDPref = new Fl_Preferences(app, "TgScopeMbxID");
	LogIDPref = new Fl_Preferences(app, "TgLogMbxID");
	Fl_Window *w;
	{ Fl_Window *o = Window = new Fl_Window(300, 250);
	  w = o;
	  o->label("Target Selection");
	  o->user_data((void *)(this));
	  { Fl_Group *o = new Fl_Group(0, 0, 300, 250);
	    { Fl_Group *o = TargetInfo = new Fl_Group(0, 0, 300, 250);
	      { Fl_Input *o = IPaddress = new Fl_Input(20, 20, 200, 20, "IP address");
		char *flexBuffer;
		o->labelsize(12);
		o->align(FL_ALIGN_TOP_LEFT);
		IPaddrPref->get("targetIP", flexBuffer, "127.0.0.1" );
		IPaddress->value(flexBuffer);
		if (flexBuffer) free(flexBuffer);
	      }
	      { Fl_Input *o = InterfaceTaskID = new Fl_Input(20, 70, 100, 20, "Interface Task Name");
		char *flexBuffer;
		o->labelsize(12);
		o->align(FL_ALIGN_TOP_LEFT);
		o->maximum_size(6);
		IfTaskIDPref->get("targetIf", flexBuffer, "IFTASK");
		InterfaceTaskID->value(flexBuffer);
		if (flexBuffer) free(flexBuffer);
	      }
	      { Fl_Input *o = ScopeMbxID = new Fl_Input(20, 120, 100, 20, "Scope Mailboxes Identifier");
		char *flexBuffer;
		o->labelsize(12);
		o->align(FL_ALIGN_TOP_LEFT);
		o->maximum_size(3);
		ScopeIDPref->get("targetScopeMbx", flexBuffer, "RTS");
		ScopeMbxID->value(flexBuffer);
		if (flexBuffer) free(flexBuffer);
	      }
	      { Fl_Input *o = LogMbxID = new Fl_Input(20, 170, 100, 20, "Log Mailboxes Identifier");
		char *flexBuffer;
		o->labelsize(12);
		o->align(FL_ALIGN_TOP_LEFT);
		o->maximum_size(3);
		LogIDPref->get("targetLogMbx", flexBuffer, "RTL");
		LogMbxID->value(flexBuffer);
		if (flexBuffer) free(flexBuffer);
	      }
	      o->end();
	    }
	    { Fl_Group *o = Buttons = new Fl_Group(0, 0, 300, 250);
	      { Fl_Button *o = Ok = new Fl_Button(140, 210, 70, 25, "OK");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_ok);
	      }
	      { Fl_Button *o = Cancel = new Fl_Button(220, 210, 70, 25, "Cancel");
		o->labelsize(12);
		o->callback((Fl_Callback *)cb_cancel);
	      }
	      o->end();
	    }
	    o->end();
          }
	}
};
